# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import logging

from subvertpy import SubversionException
from tenacity import retry
from tenacity.before_sleep import before_sleep_log
from tenacity.retry import retry_if_exception
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_exponential

logger = logging.getLogger(__name__)

SVN_RETRY_WAIT_EXP_BASE = 10
SVN_RETRY_MAX_ATTEMPTS = 5


def is_retryable_svn_exception(exception):
    if isinstance(exception, SubversionException):
        return exception.args[0].startswith(
            (
                "Connection timed out",
                "Unable to connect to a repository at URL",
                "Error running context: The server unexpectedly closed the connection",
                "ra_serf: The server sent a truncated HTTP response body",
            )
        )
    return isinstance(exception, (ConnectionResetError, TimeoutError))


def svn_retry():
    return retry(
        retry=retry_if_exception(is_retryable_svn_exception),
        wait=wait_exponential(exp_base=SVN_RETRY_WAIT_EXP_BASE),
        stop=stop_after_attempt(max_attempt_number=SVN_RETRY_MAX_ATTEMPTS),
        before_sleep=before_sleep_log(logger, logging.DEBUG),
        reraise=True,
    )



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import uuid

import pytest

from swh.scheduler.model import ListedOrigin, Lister

NAMESPACE = "swh.loader.svn"


@pytest.fixture
def svn_lister():
    return Lister(name="svn-lister", instance_name="example", id=uuid.uuid4())


@pytest.fixture
def svn_listed_origin(svn_lister):
    return ListedOrigin(
        lister_id=svn_lister.id, url="svn://example.org/repo", visit_type="svn"
    )


@pytest.mark.parametrize("extra_loader_arguments", [{}, {"visit_date": "now"}])
def test_svn_loader_for_listed_origin(
    loading_task_creation_for_listed_origin_test,
    svn_lister,
    svn_listed_origin,
    extra_loader_arguments,
):
    svn_listed_origin.extra_loader_arguments = extra_loader_arguments

    loading_task_creation_for_listed_origin_test(
        loader_class_name=f"{NAMESPACE}.loader.SvnLoader",
        task_function_name=f"{NAMESPACE}.tasks.LoadSvnRepository",
        lister=svn_lister,
        listed_origin=svn_listed_origin,
    )


@pytest.mark.parametrize(
    "extra_loader_arguments",
    [{"archive_path": "some-path"}, {"archive_path": "some-path", "visit_date": "now"}],
)
def test_svn_loader_from_dump_for_listed_origin(
    loading_task_creation_for_listed_origin_test,
    svn_lister,
    svn_listed_origin,
    extra_loader_arguments,
):
    svn_listed_origin.extra_loader_arguments = extra_loader_arguments

    loading_task_creation_for_listed_origin_test(
        loader_class_name=f"{NAMESPACE}.loader.SvnLoaderFromDumpArchive",
        task_function_name=f"{NAMESPACE}.tasks.MountAndLoadSvnRepository",
        lister=svn_lister,
        listed_origin=svn_listed_origin,
    )


@pytest.mark.parametrize("extra_loader_arguments", [{}, {"visit_date": "now"}])
def test_svn_loader_from_remote_dump_for_listed_origin(
    loading_task_creation_for_listed_origin_test,
    svn_lister,
    svn_listed_origin,
    extra_loader_arguments,
):
    svn_listed_origin.extra_loader_arguments = extra_loader_arguments

    loading_task_creation_for_listed_origin_test(
        loader_class_name=f"{NAMESPACE}.loader.SvnLoaderFromRemoteDump",
        task_function_name=f"{NAMESPACE}.tasks.DumpMountAndLoadSvnRepository",
        lister=svn_lister,
        listed_origin=svn_listed_origin,
    )



# Copyright (C) 2016-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Remote Access client to svn server.

"""

from __future__ import annotations

import codecs
from collections import defaultdict
from dataclasses import dataclass, field
from distutils.dir_util import copy_tree
from itertools import chain
import logging
import os
import shutil
import tempfile
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

import click
from subvertpy import SubversionException, delta, properties
from subvertpy.ra import Auth, RemoteAccess, get_username_provider

from swh.model import from_disk, hashutil
from swh.model.from_disk import DiskBackedContent
from swh.model.model import Content, Directory, SkippedContent

if TYPE_CHECKING:
    from swh.loader.svn.svn import SvnRepo

from swh.loader.svn.utils import (
    is_recursive_external,
    parse_external_definition,
    svn_urljoin,
)

_eol_style = {"native": b"\n", "CRLF": b"\r\n", "LF": b"\n", "CR": b"\r"}

logger = logging.getLogger(__name__)


def _normalize_line_endings(lines: bytes, eol_style: str = "native") -> bytes:
    r"""Normalize line endings to unix (\\n), windows (\\r\\n) or mac (\\r).

    Args:
        lines: The lines to normalize

        eol_style: The line ending format as defined for
            svn:eol-style property. Acceptable values are 'native',
            'CRLF', 'LF' and 'CR'

    Returns:
        Lines with endings normalized
    """
    if eol_style in _eol_style:
        lines = lines.replace(_eol_style["CRLF"], _eol_style["LF"]).replace(
            _eol_style["CR"], _eol_style["LF"]
        )
        if _eol_style[eol_style] != _eol_style["LF"]:
            lines = lines.replace(_eol_style["LF"], _eol_style[eol_style])

    return lines


def apply_txdelta_handler(
    sbuf: bytes, target_stream: BinaryIO
) -> Callable[[Any, bytes, BinaryIO], None]:
    """Return a function that can be called repeatedly with txdelta windows.
    When done, closes the target_stream.

    Adapted from subvertpy.delta.apply_txdelta_handler to close the
    stream when done.

    Args:
        sbuf: Source buffer
        target_stream: Target stream to write to.

    Returns:
        Function to be called to apply txdelta windows

    """

    def apply_window(
        window: Any, sbuf: bytes = sbuf, target_stream: BinaryIO = target_stream
    ):
        if window is None:
            target_stream.close()
            return  # Last call
        patch = delta.apply_txdelta_window(sbuf, window)
        target_stream.write(patch)

    return apply_window


def read_svn_link(data: bytes) -> Tuple[bytes, bytes]:
    """Read the svn link's content.

    Args:
        data: svn link's raw content

    Returns:
        The tuple of (filetype, destination path)

    """
    split_byte = b" "
    first_line = data.split(b"\n")[0]
    filetype, *src = first_line.split(split_byte)
    target = split_byte.join(src)
    return filetype, target


def is_file_an_svnlink_p(fullpath: bytes) -> Tuple[bool, bytes]:
    """Determine if a filepath is an svnlink or something else.

    Args:
        fullpath: Full path to the potential symlink to check

    Returns:
        Tuple containing a boolean value to determine if it's indeed a symlink
        (as per svn) and the link target.

    """
    if os.path.islink(fullpath):
        return False, b""
    with open(fullpath, "rb") as f:
        filetype, src = read_svn_link(f.read())
        return filetype == b"link", src


def _ra_codecs_error_handler(e: UnicodeError) -> Tuple[Union[str, bytes], int]:
    """Subvertpy may fail to decode to utf-8 the user svn properties.  As
       they are not used by the loader, return an empty string instead
       of the decoded content.

    Args:
        e: exception raised during the svn properties decoding.

    """
    return "", cast(UnicodeDecodeError, e).end


DEFAULT_FLAG = 0
EXEC_FLAG = 1
NOEXEC_FLAG = 2

SVN_PROPERTY_EOL = "svn:eol-style"


@dataclass
class FileState:
    """Persists some file states (eg. end of lines style) across revisions while
    replaying them."""

    eol_style: Optional[str] = None
    """EOL state check mess"""

    svn_special_path_non_link_data: Optional[bytes] = None
    """keep track of non link file content with svn:special property set"""

    # default value: 0, 1: set the flag, 2: remove the exec flag
    executable: int = DEFAULT_FLAG
    """keep track if file is executable when setting svn:executable property"""

    link: bool = False
    """keep track if file is a svn link when setting svn:special property"""


class FileEditor:
    """File Editor in charge of updating file on disk and memory objects."""

    __slots__ = [
        "directory",
        "path",
        "fullpath",
        "executable",
        "link",
        "state",
        "svnrepo",
        "editor",
    ]

    def __init__(
        self,
        directory: from_disk.Directory,
        rootpath: bytes,
        path: bytes,
        state: FileState,
        svnrepo: SvnRepo,
    ):
        self.directory = directory
        self.path = path
        self.fullpath = os.path.join(rootpath, path)
        self.state = state
        self.svnrepo = svnrepo
        self.editor = svnrepo.swhreplay.editor

    def change_prop(self, key: str, value: str) -> None:
        if key == properties.PROP_EXECUTABLE:
            if value is None:  # bit flip off
                self.state.executable = NOEXEC_FLAG
            else:
                self.state.executable = EXEC_FLAG
        elif key == properties.PROP_SPECIAL:
            # Possibly a symbolic link. We cannot check further at
            # that moment though, patch(s) not being applied yet
            self.state.link = value is not None
        elif key == SVN_PROPERTY_EOL:
            # backup end of line style for file
            self.state.eol_style = value

    def __make_symlink(self, src: bytes) -> None:
        """Convert the svnlink to a symlink on disk.

        This function expects self.fullpath to be a svn link.

        Args:
            src: Path to the link's source

        Return:
            tuple: The svnlink's data tuple:

                - type (should be only 'link')
                - <path-to-src>

        """
        os.remove(self.fullpath)
        os.symlink(src=src, dst=self.fullpath)

    def __make_svnlink(self) -> bytes:
        """Convert the symlink to a svnlink on disk.

        Return:
            The symlink's svnlink data (``b'type <path-to-src>'``)

        """
        # we replace the symlink by a svnlink
        # to be able to patch the file on future commits
        src = os.readlink(self.fullpath)
        os.remove(self.fullpath)
        sbuf = b"link " + src
        with open(self.fullpath, "wb") as f:
            f.write(sbuf)
        return sbuf

    def apply_textdelta(self, base_checksum) -> Callable[[Any, bytes, BinaryIO], None]:
        # if the filepath matches an external, do not apply local patch
        if self.path in self.editor.external_paths:
            return lambda *args: None

        if os.path.lexists(self.fullpath):
            if os.path.islink(self.fullpath):
                # svn does not deal with symlink so we transform into
                # real svn symlink for potential patching in later
                # commits
                sbuf = self.__make_svnlink()
                self.state.link = True
            else:
                with open(self.fullpath, "rb") as f:
                    sbuf = f.read()
        else:
            sbuf = b""

        t = open(self.fullpath, "wb")
        return apply_txdelta_handler(sbuf, target_stream=t)

    def close(self) -> None:
        """When done with the file, this is called.

        So the file exists and is updated, we can:

        - adapt accordingly its execution flag if any
        - compute the objects' checksums
        - replace the svnlink with a real symlink (for disk
          computation purposes)

        """

        if self.state.link:
            # can only check now that the link is a real one
            # since patch has been applied
            is_link, src = is_file_an_svnlink_p(self.fullpath)
            if is_link:
                self.__make_symlink(src)
            elif not os.path.isdir(self.fullpath):  # not a real link ...
                # when a file with the svn:special property set is not a svn link,
                # the svn export operation might extract a truncated version of it
                # if it is a binary file, so ensure to produce the same file as the
                # export operation.
                with open(self.fullpath, "rb") as f:
                    content = f.read()
                self.svnrepo.export(
                    os.path.join(self.svnrepo.remote_url, os.fsdecode(self.path)),
                    to=self.fullpath,
                    peg_rev=self.editor.revnum,
                    ignore_keywords=True,
                    overwrite=True,
                )
                with open(self.fullpath, "rb") as f:
                    exported_data = f.read()
                    if exported_data != content:
                        # keep track of original file content in order to restore
                        # it if the svn:special property gets unset in another revision
                        self.state.svn_special_path_non_link_data = content
        elif os.path.islink(self.fullpath):
            # path was a symbolic link in previous revision but got the property
            # svn:special unset in current one, revert its content to svn link format
            self.__make_svnlink()
        elif self.state.svn_special_path_non_link_data is not None:
            # path was a non link file with the svn:special property previously set
            # and got truncated on export, restore its original content
            with open(self.fullpath, "wb") as f:
                f.write(self.state.svn_special_path_non_link_data)
                self.state.svn_special_path_non_link_data = None

        is_link = os.path.islink(self.fullpath)
        if not is_link:  # if a link, do nothing regarding flag
            if self.state.executable == EXEC_FLAG:
                os.chmod(self.fullpath, 0o755)
            elif self.state.executable == NOEXEC_FLAG:
                os.chmod(self.fullpath, 0o644)

        # And now compute file's checksums
        if self.state.eol_style and not is_link:
            # ensure to normalize line endings as defined by svn:eol-style
            # property to get the same file checksum as after an export
            # or checkout operation with subversion
            with open(self.fullpath, "rb") as f:
                data = f.read()
                data = _normalize_line_endings(data, self.state.eol_style)
                mode = os.lstat(self.fullpath).st_mode
                self.directory[self.path] = from_disk.Content.from_bytes(
                    mode=mode, data=data
                )
        else:
            self.directory[self.path] = from_disk.Content.from_file(path=self.fullpath)


ExternalDefinition = Tuple[str, Optional[int], bool]


@dataclass
class DirState:
    """Persists some directory states (eg. externals) across revisions while
    replaying them."""

    externals: Dict[str, List[ExternalDefinition]] = field(default_factory=dict)
    """Map a path in the directory to a list of (external_url, revision, relative_url)
    targeting it"""
    externals_paths: Set[bytes] = field(default_factory=set)
    """Keep track of all external paths reachable from the directory"""


class DirEditor:
    """Directory Editor in charge of updating directory hashes computation.

    This implementation includes empty folder in the hash computation.

    """

    __slots__ = [
        "directory",
        "rootpath",
        "path",
        "file_states",
        "dir_states",
        "svnrepo",
        "editor",
        "externals",
    ]

    def __init__(
        self,
        directory: from_disk.Directory,
        rootpath: bytes,
        path: bytes,
        file_states: Dict[bytes, FileState],
        dir_states: Dict[bytes, DirState],
        svnrepo: SvnRepo,
    ):
        self.directory = directory
        self.rootpath = rootpath
        self.path = path
        # build directory on init
        os.makedirs(rootpath, exist_ok=True)
        self.file_states = file_states
        self.dir_states = dir_states
        self.svnrepo = svnrepo
        self.editor = svnrepo.swhreplay.editor
        self.externals: Dict[str, List[ExternalDefinition]] = {}

    def remove_child(self, path: bytes) -> None:
        """Remove a path from the current objects.

        The path can be resolved as link, file or directory.

        This function takes also care of removing the link between the
        child and the parent.

        Args:
            path: to remove from the current objects.

        """
        if path in self.directory:
            entry_removed = self.directory[path]
            del self.directory[path]
            fpath = os.path.join(self.rootpath, path)
            if isinstance(entry_removed, from_disk.Directory):
                shutil.rmtree(fpath)
            else:
                os.remove(fpath)

        # when deleting a directory ensure to remove any svn property for the
        # file it contains as they can be added again later in another revision
        # without the same property set
        fullpath = os.path.join(self.rootpath, path)
        for state_path in list(self.file_states):
            if state_path.startswith(fullpath + b"/"):
                del self.file_states[state_path]

    def open_directory(self, path: str, *args) -> DirEditor:
        """Updating existing directory."""
        return DirEditor(
            self.directory,
            rootpath=self.rootpath,
            path=os.fsencode(path),
            file_states=self.file_states,
            dir_states=self.dir_states,
            svnrepo=self.svnrepo,
        )

    def add_directory(
        self, path: str, copyfrom_path: Optional[str] = None, copyfrom_rev: int = -1
    ) -> DirEditor:
        """Adding a new directory."""
        path_bytes = os.fsencode(path)
        fullpath = os.path.join(self.rootpath, path_bytes)

        os.makedirs(fullpath, exist_ok=True)
        if copyfrom_rev == -1:
            if path_bytes and path_bytes not in self.directory:
                self.dir_states[path_bytes] = DirState()
                self.directory[path_bytes] = from_disk.Directory()
        else:
            url = svn_urljoin(self.svnrepo.remote_url, copyfrom_path)
            self.remove_child(path_bytes)
            self.svnrepo.export(
                url,
                to=fullpath,
                peg_rev=copyfrom_rev,
                ignore_keywords=True,
                overwrite=True,
                ignore_externals=True,
            )
            self.directory[path_bytes] = from_disk.Directory.from_disk(path=fullpath)

        return DirEditor(
            self.directory,
            self.rootpath,
            path_bytes,
            self.file_states,
            self.dir_states,
            svnrepo=self.svnrepo,
        )

    def open_file(self, path: str, *args) -> FileEditor:
        """Updating existing file."""
        path_bytes = os.fsencode(path)
        self.directory[path_bytes] = from_disk.Content()
        fullpath = os.path.join(self.rootpath, path_bytes)
        return FileEditor(
            self.directory,
            rootpath=self.rootpath,
            path=path_bytes,
            state=self.file_states[fullpath],
            svnrepo=self.svnrepo,
        )

    def add_file(
        self, path: str, copyfrom_path: Optional[str] = None, copyfrom_rev: int = -1
    ) -> FileEditor:
        """Creating a new file."""
        path_bytes = os.fsencode(path)
        fullpath = os.path.join(self.rootpath, path_bytes)

        self.file_states[fullpath] = FileState()
        if copyfrom_rev == -1:
            self.directory[path_bytes] = from_disk.Content()
        else:
            url = svn_urljoin(self.svnrepo.remote_url, copyfrom_path)
            self.remove_child(path_bytes)
            self.svnrepo.export(
                url,
                to=fullpath,
                peg_rev=copyfrom_rev,
                ignore_keywords=True,
                overwrite=True,
            )
            self.directory[path_bytes] = from_disk.Content.from_file(path=fullpath)

        return FileEditor(
            self.directory,
            self.rootpath,
            path_bytes,
            state=self.file_states[fullpath],
            svnrepo=self.svnrepo,
        )

    def change_prop(self, key: str, value: str) -> None:
        """Change property callback on directory."""
        if key == properties.PROP_EXTERNALS:
            logger.debug(
                "Setting '%s' property with value '%s' on path %s",
                key,
                value,
                self.path,
            )
            self.externals = defaultdict(list)
            if value is not None:
                try:
                    # externals are set on that directory path, parse and store them
                    # for later processing in the close method
                    for external in value.split("\n"):
                        external = external.rstrip("\r")
                        # skip empty line or comment
                        if not external or external.startswith("#"):
                            continue
                        (
                            path,
                            external_url,
                            revision,
                            relative_url,
                        ) = parse_external_definition(
                            external, os.fsdecode(self.path), self.svnrepo.origin_url
                        )
                        self.externals[path].append(
                            (external_url, revision, relative_url)
                        )
                except ValueError:
                    logger.debug(
                        "Failed to parse external: %s\n"
                        "Externals defined on path %s will not be processed",
                        external,
                        self.path,
                    )
                    # as the official subversion client, do not process externals in case
                    # of parsing error
                    self.externals = {}

            if not self.externals:
                # externals might have been unset on that directory path,
                # remove associated paths from the reconstructed filesystem
                externals = self.dir_states[self.path].externals
                for path in externals.keys():
                    self.remove_external_path(os.fsencode(path))

                self.dir_states[self.path].externals = {}

    def delete_entry(self, path: str, revision: int) -> None:
        """Remove a path."""
        path_bytes = os.fsencode(path)
        fullpath = os.path.join(self.rootpath, path_bytes)

        if os.path.isdir(fullpath):
            # remove all external paths associated to the removed directory
            # (we cannot simply remove a root external directory as externals
            # paths associated to ancestor directories can overlap)
            for external_path in self.dir_states[path_bytes].externals_paths:
                self.remove_external_path(
                    external_path,
                    root_path=path_bytes,
                    remove_subpaths=False,
                    force=True,
                )

        if path_bytes not in self.editor.external_paths:
            self.file_states.pop(fullpath, None)
            self.remove_child(path_bytes)
        elif os.path.isdir(fullpath):
            # versioned and external paths can overlap so we need to iterate on
            # all subpaths to check which ones to remove
            for root, dirs, files in os.walk(fullpath):
                for p in chain(dirs, files):
                    full_repo_path = os.path.join(root, p)
                    repo_path = full_repo_path.replace(self.rootpath + b"/", b"")
                    if repo_path not in self.editor.external_paths:
                        self.file_states.pop(full_repo_path, None)
                        self.remove_child(repo_path)

    def close(self):
        """Function called when we finish processing a repository.

        SVN external definitions are processed by it.
        """

        prev_externals = self.dir_states[self.path].externals

        if self.externals:
            # externals definition list might have changed in the current replayed
            # revision, we need to determine if some were removed and delete the
            # associated paths
            externals = self.externals
            prev_externals_set = {
                (path, url, rev)
                for path in prev_externals.keys()
                for (url, rev, _) in prev_externals[path]
            }
            externals_set = {
                (path, url, rev)
                for path in externals.keys()
                for (url, rev, _) in externals[path]
            }
            old_externals = prev_externals_set - externals_set
            for path, _, _ in old_externals:
                self.remove_external_path(os.fsencode(path))
        else:
            # some external paths might have been removed in the current replayed
            # revision by a delete operation on an overlapping versioned path so we
            # need to restore them
            externals = prev_externals

        # For each external, try to export it in reconstructed filesystem
        for path, externals_def in externals.items():
            for i, external in enumerate(externals_def):
                external_url, revision, relative_url = external
                self.process_external(
                    path,
                    external_url,
                    revision,
                    relative_url,
                    remove_target_path=i == 0,
                )

        # backup externals in directory state
        if self.externals:
            self.dir_states[self.path].externals = self.externals

        # do operations below only when closing the root directory
        if self.path == b"":
            self.svnrepo.has_relative_externals = any(
                relative_url
                for (_, relative_url) in self.editor.valid_externals.values()
            )

            self.svnrepo.has_recursive_externals = any(
                is_recursive_external(
                    self.svnrepo.origin_url,
                    os.fsdecode(path),
                    external_path,
                    external_url,
                )
                for path, dir_state in self.dir_states.items()
                for external_path in dir_state.externals.keys()
                for (external_url, _, _) in dir_state.externals[external_path]
            )
            if self.svnrepo.has_recursive_externals:
                # If the repository has recursive externals, we stop processing
                # externals and remove those already exported,
                # We will then ignore externals when exporting the revision to
                # check for divergence with the reconstructed filesystem.
                for external_path in list(self.editor.external_paths):
                    self.remove_external_path(external_path, force=True)

    def process_external(
        self,
        path: str,
        external_url: str,
        revision: Optional[int],
        relative_url: bool,
        remove_target_path: bool = True,
    ) -> None:
        external = (external_url, revision, relative_url)
        dest_path = os.fsencode(path)
        dest_fullpath = os.path.join(self.path, dest_path)
        prev_externals = self.dir_states[self.path].externals
        if (
            path in prev_externals
            and external in prev_externals[path]
            and dest_fullpath in self.directory
        ):
            # external already exported, nothing to do
            return

        if is_recursive_external(
            self.svnrepo.origin_url, os.fsdecode(self.path), path, external_url
        ):
            # recursive external, skip it
            return

        logger.debug(
            "Exporting external %s%s to path %s",
            external_url,
            f"@{revision}" if revision else "",
            dest_fullpath,
        )

        if external not in self.editor.externals_cache:

            try:
                # try to export external in a temporary path, destination path could
                # be versioned and must be overridden only if the external URL is
                # still valid
                temp_dir = os.fsencode(
                    tempfile.mkdtemp(dir=self.editor.externals_cache_dir)
                )
                temp_path = os.path.join(temp_dir, dest_path)
                os.makedirs(b"/".join(temp_path.split(b"/")[:-1]), exist_ok=True)
                if external_url not in self.editor.dead_externals:
                    url = external_url.rstrip("/")
                    origin_url = self.svnrepo.origin_url.rstrip("/")
                    if (
                        url.startswith(origin_url + "/")
                        and not self.svnrepo.has_relative_externals
                    ):
                        url = url.replace(origin_url, self.svnrepo.remote_url)
                    self.svnrepo.export(
                        url,
                        to=temp_path,
                        peg_rev=revision,
                        ignore_keywords=True,
                    )
                    self.editor.externals_cache[external] = temp_path

            except SubversionException as se:
                # external no longer available (404)
                logger.debug(se)
                self.editor.dead_externals.add(external_url)

        else:
            temp_path = self.editor.externals_cache[external]

        # subversion export will always create the subdirectories of the external
        # path regardless the validity of the remote URL
        dest_path_split = dest_path.split(b"/")
        current_path = self.path
        self.add_directory(os.fsdecode(current_path))
        for subpath in dest_path_split[:-1]:
            current_path = os.path.join(current_path, subpath)
            self.add_directory(os.fsdecode(current_path))

        if os.path.exists(temp_path):
            # external successfully exported

            if remove_target_path:
                # remove previous path in from_disk model
                self.remove_external_path(dest_path, remove_subpaths=False)

            # mark external as valid
            self.editor.valid_externals[dest_fullpath] = (
                external_url,
                relative_url,
            )

            # copy exported path to reconstructed filesystem
            fullpath = os.path.join(self.rootpath, dest_fullpath)

            if os.path.isfile(temp_path):
                if os.path.islink(fullpath):
                    # remove destination file if it is a link
                    os.remove(fullpath)
                shutil.copy(os.fsdecode(temp_path), os.fsdecode(fullpath))
                self.directory[dest_fullpath] = from_disk.Content.from_file(
                    path=fullpath
                )
            else:
                self.add_directory(os.fsdecode(dest_fullpath))

                # copy_tree needs sub-directories to exist in destination
                for root, dirs, files in os.walk(temp_path):
                    for dir in dirs:
                        temp_dir_fullpath = os.path.join(root, dir)
                        if os.path.islink(temp_dir_fullpath):
                            # do not create folder if it's a link or copy_tree will fail
                            continue
                        subdir = temp_dir_fullpath.replace(temp_path + b"/", b"")
                        self.add_directory(
                            os.fsdecode(os.path.join(dest_fullpath, subdir))
                        )

                copy_tree(
                    os.fsdecode(temp_path),
                    os.fsdecode(fullpath),
                    preserve_symlinks=True,
                )

                # TODO: replace code above by the line below once we use Python >= 3.8 in production # noqa
                # shutil.copytree(temp_path, fullpath, symlinks=True, dirs_exist_ok=True) # noqa

                self.directory[dest_fullpath] = from_disk.Directory.from_disk(
                    path=fullpath
                )

            # update set of external paths reachable from the directory
            external_paths = set()
            dest_path_part = dest_path.split(b"/")
            for i in range(1, len(dest_path_part) + 1):
                external_paths.add(b"/".join(dest_path_part[:i]))

            for root, dirs, files in os.walk(temp_path):
                external_paths.update(
                    [
                        os.path.join(
                            dest_path,
                            os.path.join(root, p).replace(temp_path, b"").strip(b"/"),
                        )
                        for p in chain(dirs, files)
                    ]
                )

            self.dir_states[self.path].externals_paths.update(external_paths)

            for external_path in external_paths:
                self.editor.external_paths[os.path.join(self.path, external_path)] += 1

            # ensure hash update for the directory with externals set
            self.directory[self.path].update_hash(force=True)

    def remove_external_path(
        self,
        external_path: bytes,
        remove_subpaths: bool = True,
        force: bool = False,
        root_path: Optional[bytes] = None,
    ) -> None:
        """Remove a previously exported SVN external path from
        the reconstructed filesystem.
        """
        path = root_path if root_path else self.path
        fullpath = os.path.join(path, external_path)

        # decrement number of references for external path when we really remove it
        # (when remove_subpaths is False, we just cleanup the external path before
        # copying exported paths in it)
        if force or (fullpath in self.editor.external_paths and remove_subpaths):
            self.editor.external_paths[fullpath] -= 1

        if (
            fullpath in self.editor.external_paths
            and self.editor.external_paths[fullpath] == 0
        ):
            self.remove_child(fullpath)
            self.editor.external_paths.pop(fullpath, None)
            self.editor.valid_externals.pop(fullpath, None)
            for path in list(self.editor.external_paths):
                if path.startswith(fullpath + b"/"):
                    self.editor.external_paths[path] -= 1
                    if self.editor.external_paths[path] == 0:
                        self.editor.external_paths.pop(path)

        if remove_subpaths:
            subpath_split = fullpath.split(b"/")[:-1]
            for i in reversed(range(1, len(subpath_split) + 1)):
                # delete external sub-directory only if it is not versioned
                subpath = b"/".join(subpath_split[0:i])
                try:
                    self.svnrepo.client.info(
                        svn_urljoin(self.svnrepo.remote_url, os.fsdecode(subpath)),
                        peg_revision=self.editor.revnum,
                        revision=self.editor.revnum,
                    )
                except SubversionException:
                    self.remove_child(subpath)
                else:
                    break

        try:
            # externals can overlap with versioned files so we must restore
            # them after removing the path above
            dest_path = os.path.join(self.rootpath, fullpath)
            self.svnrepo.client.export(
                svn_urljoin(self.svnrepo.remote_url, os.fsdecode(fullpath)),
                to=dest_path,
                peg_rev=self.editor.revnum,
                ignore_keywords=True,
            )
            if os.path.isfile(dest_path) or os.path.islink(dest_path):
                self.directory[fullpath] = from_disk.Content.from_file(path=dest_path)
            else:
                self.directory[fullpath] = from_disk.Directory.from_disk(path=dest_path)
        except SubversionException:
            pass


class Editor:
    """Editor in charge of replaying svn events and computing objects
    along.

    This implementation accounts for empty folder during hash
    computations.

    """

    def __init__(
        self,
        rootpath: bytes,
        directory: from_disk.Directory,
        svnrepo: SvnRepo,
        temp_dir: str,
    ):
        self.rootpath = rootpath
        self.directory = directory
        self.file_states: Dict[bytes, FileState] = defaultdict(FileState)
        self.dir_states: Dict[bytes, DirState] = defaultdict(DirState)
        self.external_paths: Dict[bytes, int] = defaultdict(int)
        self.valid_externals: Dict[bytes, Tuple[str, bool]] = {}
        self.dead_externals: Set[str] = set()
        self.externals_cache_dir = tempfile.mkdtemp(dir=temp_dir)
        self.externals_cache: Dict[ExternalDefinition, bytes] = {}
        self.svnrepo = svnrepo
        self.revnum = None

    def set_target_revision(self, revnum) -> None:
        self.revnum = revnum

    def abort(self) -> None:
        pass

    def close(self) -> None:
        pass

    def open_root(self, base_revnum: int) -> DirEditor:
        return DirEditor(
            self.directory,
            rootpath=self.rootpath,
            path=b"",
            file_states=self.file_states,
            dir_states=self.dir_states,
            svnrepo=self.svnrepo,
        )


class Replay:
    """Replay class."""

    def __init__(
        self,
        conn: RemoteAccess,
        rootpath: bytes,
        svnrepo: SvnRepo,
        temp_dir: str,
        directory: Optional[from_disk.Directory] = None,
    ):
        self.conn = conn
        self.rootpath = rootpath
        if directory is None:
            directory = from_disk.Directory()
        self.directory = directory
        self.editor = Editor(
            rootpath=rootpath, directory=directory, svnrepo=svnrepo, temp_dir=temp_dir
        )

    def replay(self, rev: int, low_water_mark: int) -> from_disk.Directory:
        """Replay svn actions between rev and rev+1.

        This method updates in place the self.editor.directory, as well as the
        filesystem.

        Returns:
           The updated root directory

        """
        codecs.register_error("strict", _ra_codecs_error_handler)
        self.conn.replay(rev, low_water_mark, self.editor)
        codecs.register_error("strict", codecs.strict_errors)
        return self.editor.directory

    def compute_objects(
        self, rev: int, low_water_mark: int
    ) -> Tuple[List[Content], List[SkippedContent], List[Directory]]:
        """Compute objects added or modified at revisions rev.
        Expects the state to be at previous revision's objects.

        Args:
            rev: The revision to start the replay from.

        Returns:
            The updated objects between rev and rev+1. Beware that this
            mutates the filesystem at rootpath accordingly.

        """
        self.replay(rev, low_water_mark)

        contents: List[Content] = []
        skipped_contents: List[SkippedContent] = []
        directories: List[Directory] = []

        for obj_node in self.directory.collect():
            obj = obj_node.to_model()  # type: ignore
            obj_type = obj.object_type
            if obj_type in (Content.object_type, DiskBackedContent.object_type):
                contents.append(obj.with_data())
            elif obj_type == SkippedContent.object_type:
                skipped_contents.append(obj)
            elif obj_type == Directory.object_type:
                directories.append(obj)
            else:
                assert False, obj_type

        return contents, skipped_contents, directories


@click.command()
@click.option("--local-url", default="/tmp", help="local svn working copy")
@click.option(
    "--svn-url",
    default="file:///home/storage/svn/repos/pkg-fox",
    help="svn repository's url.",
)
@click.option(
    "--revision-start",
    default=1,
    type=click.INT,
    help="svn repository's starting revision.",
)
@click.option(
    "--revision-end",
    default=-1,
    type=click.INT,
    help="svn repository's ending revision.",
)
@click.option(
    "--debug/--nodebug",
    default=True,
    help="Indicates if the server should run in debug mode.",
)
@click.option(
    "--cleanup/--nocleanup",
    default=True,
    help="Indicates whether to cleanup disk when done or not.",
)
def main(local_url, svn_url, revision_start, revision_end, debug, cleanup):
    """Script to present how to use Replay class."""
    conn = RemoteAccess(svn_url.encode("utf-8"), auth=Auth([get_username_provider()]))

    os.makedirs(local_url, exist_ok=True)

    rootpath = tempfile.mkdtemp(
        prefix=local_url, suffix="-" + os.path.basename(svn_url)
    )

    rootpath = os.fsencode(rootpath)

    # Do not go beyond the repository's latest revision
    revision_end_max = conn.get_latest_revnum()
    if revision_end == -1:
        revision_end = revision_end_max

    revision_end = min(revision_end, revision_end_max)

    try:
        replay = Replay(conn, rootpath)

        for rev in range(revision_start, revision_end + 1):
            contents, skipped_contents, directories = replay.compute_objects(rev)
            print(
                "r%s %s (%s new contents, %s new directories)"
                % (
                    rev,
                    hashutil.hash_to_hex(replay.directory.hash),
                    len(contents) + len(skipped_contents),
                    len(directories),
                )
            )

        if debug:
            print("%s" % rootpath.decode("utf-8"))
    finally:
        if cleanup:
            if os.path.exists(rootpath):
                shutil.rmtree(rootpath)


if __name__ == "__main__":
    main()



# Copyright (C) 2016-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os
from pathlib import Path
import pty
import re
import shutil
from subprocess import Popen, run

import pytest

from swh.loader.svn import utils
from swh.loader.tests import prepare_repository_from_archive


def test_outputstream():
    stdout_r, stdout_w = pty.openpty()
    echo = Popen(["echo", "-e", "foo\nbar\nbaz"], stdout=stdout_w)
    os.close(stdout_w)
    stdout_stream = utils.OutputStream(stdout_r)
    lines = []
    while True:
        current_lines, readable = stdout_stream.read_lines()
        lines += current_lines
        if not readable:
            break
    echo.wait()
    os.close(stdout_r)
    assert lines == ["foo", "bar", "baz"]


def test_init_svn_repo_from_dump(datadir, tmp_path):
    """Mounting svn repository out of a dump is ok"""
    dump_name = "penguinsdbtools2018.dump.gz"
    dump_path = os.path.join(datadir, dump_name)

    tmp_repo, repo_path = utils.init_svn_repo_from_dump(
        dump_path, gzip=True, cleanup_dump=False, root_dir=tmp_path
    )

    assert os.path.exists(dump_path), "Dump path should still exists"
    assert os.path.exists(repo_path), "Repository should exists"


def test_init_svn_repo_from_dump_svnadmin_error(tmp_path):
    """svnadmin load error should be reported in exception text"""
    dump_path = os.path.join(tmp_path, "foo")
    Path(dump_path).touch()

    with pytest.raises(
        ValueError,
        match="svnadmin: E200003: Premature end of content data in dumpstream",
    ):
        utils.init_svn_repo_from_dump(dump_path, cleanup_dump=False, root_dir=tmp_path)


def test_init_svn_repo_from_dump_and_cleanup(datadir, tmp_path):
    """Mounting svn repository with a dump cleanup after is ok"""
    dump_name = "penguinsdbtools2018.dump.gz"
    dump_ori_path = os.path.join(datadir, dump_name)

    dump_path = os.path.join(tmp_path, dump_name)
    shutil.copyfile(dump_ori_path, dump_path)

    assert os.path.exists(dump_path)
    assert os.path.exists(dump_ori_path)

    tmp_repo, repo_path = utils.init_svn_repo_from_dump(
        dump_path, gzip=True, root_dir=tmp_path
    )

    assert not os.path.exists(dump_path), "Dump path should no longer exists"
    assert os.path.exists(repo_path), "Repository should exists"
    assert os.path.exists(dump_ori_path), "Original dump path should still exists"


def test_init_svn_repo_from_dump_and_cleanup_already_done(
    datadir, tmp_path, mocker, caplog
):
    """Mounting svn repository out of a dump is ok"""
    caplog.set_level(logging.INFO, "swh.loader.svn.utils")

    dump_name = "penguinsdbtools2018.dump.gz"
    dump_ori_path = os.path.join(datadir, dump_name)

    mock_remove = mocker.patch("os.remove")
    mock_remove.side_effect = FileNotFoundError

    dump_path = os.path.join(tmp_path, dump_name)
    shutil.copyfile(dump_ori_path, dump_path)

    assert os.path.exists(dump_path)
    assert os.path.exists(dump_ori_path)

    tmp_repo, repo_path = utils.init_svn_repo_from_dump(
        dump_path, gzip=True, root_dir=tmp_path
    )

    assert os.path.exists(repo_path), "Repository should exists"
    assert os.path.exists(dump_ori_path), "Original dump path should still exists"

    assert len(caplog.record_tuples) == 1
    assert "Failure to remove" in caplog.record_tuples[0][2]
    assert mock_remove.called


def test_init_svn_repo_from_truncated_dump(datadir, tmp_path):
    """Mounting partial svn repository from a truncated dump should work"""

    # prepare a repository
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    # dump it to file
    dump_path = str(tmp_path / f"{archive_name}.dump")
    truncated_dump_path = str(tmp_path / f"{archive_name}_truncated.dump")
    svnrdump_cmd = ["svnrdump", "dump", repo_url]
    with open(dump_path, "wb") as dump:
        run(svnrdump_cmd, stdout=dump)

    # create a truncated dump file that will generate a "svnadmin load" error
    with open(dump_path, "rb") as dump, open(
        truncated_dump_path, "wb"
    ) as truncated_dump:
        dump_lines = dump.readlines()
        assert len(dump_lines) > 150
        truncated_dump_content = b"".join(dump_lines[:150])
        truncated_dump.write(truncated_dump_content)

        # compute max revision number with non truncated data
        revs = re.findall(rb"Revision-number: ([0-9]+)", truncated_dump_content)
        max_rev = int(revs[-1]) - 1

    # prepare repository from truncated dump
    _, repo_path = utils.init_svn_repo_from_dump(
        truncated_dump_path, gzip=False, root_dir=tmp_path, max_rev=max_rev
    )

    # check expected number of revisions have been loaded
    svnadmin_info = run(["svnadmin", "info", repo_path], capture_output=True, text=True)
    assert f"Revisions: {max_rev}\n" in svnadmin_info.stdout


def test_init_svn_repo_from_archive_dump(datadir, tmp_path):
    """Mounting svn repository out of an archive dump is ok"""
    dump_name = "penguinsdbtools2018.dump.gz"
    dump_path = os.path.join(datadir, dump_name)

    tmp_repo, repo_path = utils.init_svn_repo_from_archive_dump(
        dump_path, cleanup_dump=False, root_dir=tmp_path
    )

    assert os.path.exists(dump_path), "Dump path should still exists"
    assert os.path.exists(repo_path), "Repository should exists"


def test_init_svn_repo_from_archive_dump_and_cleanup(datadir, tmp_path):
    """Mounting svn repository out of a dump is ok"""
    dump_name = "penguinsdbtools2018.dump.gz"
    dump_ori_path = os.path.join(datadir, dump_name)

    dump_path = os.path.join(tmp_path, dump_name)
    shutil.copyfile(dump_ori_path, dump_path)

    assert os.path.exists(dump_path)
    assert os.path.exists(dump_ori_path)

    tmp_repo, repo_path = utils.init_svn_repo_from_archive_dump(
        dump_path, root_dir=tmp_path
    )

    assert not os.path.exists(dump_path), "Dump path should no longer exists"
    assert os.path.exists(repo_path), "Repository should exists"
    assert os.path.exists(dump_ori_path), "Original dump path should still exists"


@pytest.mark.parametrize(
    "base_url, paths_to_join, expected_result",
    [
        (
            "https://svn.example.org",
            ["repos", "test"],
            "https://svn.example.org/repos/test",
        ),
        (
            "https://svn.example.org/",
            ["repos", "test"],
            "https://svn.example.org/repos/test",
        ),
        (
            "https://svn.example.org/foo",
            ["repos", "test"],
            "https://svn.example.org/foo/repos/test",
        ),
        (
            "https://svn.example.org/foo/",
            ["/repos", "test/"],
            "https://svn.example.org/foo/repos/test",
        ),
        (
            "https://svn.example.org/foo",
            ["../bar"],
            "https://svn.example.org/bar",
        ),
    ],
)
def test_svn_urljoin(base_url, paths_to_join, expected_result):
    assert utils.svn_urljoin(base_url, *paths_to_join) == expected_result


@pytest.mark.parametrize(
    "external, dir_path, repo_url, expected_result",
    [
        # subversion < 1.5
        (
            "third-party/sounds             http://svn.example.com/repos/sounds",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("third-party/sounds", "http://svn.example.com/repos/sounds", None, False),
        ),
        (
            "third-party/skins -r148        http://svn.example.com/skinproj",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("third-party/skins", "http://svn.example.com/skinproj", 148, False),
        ),
        (
            "third-party/skins/toolkit -r21 http://svn.example.com/skin-maker",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                "third-party/skins/toolkit",
                "http://svn.example.com/skin-maker",
                21,
                False,
            ),
        ),
        # subversion >= 1.5
        (
            "      http://svn.example.com/repos/sounds third-party/sounds",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("third-party/sounds", "http://svn.example.com/repos/sounds", None, False),
        ),
        (
            "-r148 http://svn.example.com/skinproj third-party/skins",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("third-party/skins", "http://svn.example.com/skinproj", 148, False),
        ),
        (
            "-r 21 http://svn.example.com/skin-maker third-party/skins/toolkit",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                "third-party/skins/toolkit",
                "http://svn.example.com/skin-maker",
                21,
                False,
            ),
        ),
        (
            "http://svn.example.com/repos/sounds third-party/sounds",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("third-party/sounds", "http://svn.example.com/repos/sounds", None, False),
        ),
        (
            "http://svn.example.com/skinproj@148 third-party/skins",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("third-party/skins", "http://svn.example.com/skinproj", 148, False),
        ),
        (
            "http://anon:anon@svn.example.com/skin-maker@21 third-party/skins/toolkit",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                "third-party/skins/toolkit",
                "http://anon:anon@svn.example.com/skin-maker",
                21,
                False,
            ),
        ),
        (
            "-r21 http://anon:anon@svn.example.com/skin-maker third-party/skins/toolkit",  # noqa
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                "third-party/skins/toolkit",
                "http://anon:anon@svn.example.com/skin-maker",
                21,
                False,
            ),
        ),
        (
            "-r21 http://anon:anon@svn.example.com/skin-maker@21 third-party/skins/toolkit",  # noqa
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                "third-party/skins/toolkit",
                "http://anon:anon@svn.example.com/skin-maker",
                21,
                False,
            ),
        ),
        # subversion >= 1.5, relative external definitions
        (
            "^/sounds third-party/sounds",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                "third-party/sounds",
                "http://svn.example.org/repos/test/sounds",
                None,
                False,
            ),
        ),
        (
            "/skinproj@148 third-party/skins",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("third-party/skins", "http://svn.example.org/skinproj", 148, True),
        ),
        (
            "//svn.example.com/skin-maker@21 third-party/skins/toolkit",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                "third-party/skins/toolkit",
                "http://svn.example.com/skin-maker",
                21,
                True,
            ),
        ),
        (
            "^/../../skin-maker@21 third-party/skins/toolkit",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                "third-party/skins/toolkit",
                "http://svn.example.org/skin-maker",
                21,
                True,
            ),
        ),
        (
            "../skins skins",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("skins", "http://svn.example.org/repos/test/trunk/skins", None, False),
        ),
        (
            "../skins skins",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("skins", "http://svn.example.org/repos/test/trunk/skins", None, False),
        ),
        # subversion >= 1.6
        (
            'http://svn.thirdparty.com/repos/My%20Project "My Project"',
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("My Project", "http://svn.thirdparty.com/repos/My%20Project", None, False),
        ),
        (
            'http://svn.thirdparty.com/repos/My%20%20%20Project "My   Project"',
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                "My   Project",
                "http://svn.thirdparty.com/repos/My%20%20%20Project",
                None,
                False,
            ),
        ),
        (
            'http://svn.thirdparty.com/repos/%22Quotes%20Too%22 \\"Quotes\\ Too\\"',
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                '"Quotes Too"',
                "http://svn.thirdparty.com/repos/%22Quotes%20Too%22",
                None,
                False,
            ),
        ),
        (
            'http://svn.thirdparty.com/repos/%22Quotes%20%20%20Too%22 \\"Quotes\\ \\ \\ Too\\"',  # noqa
            "trunk/externals",
            "http://svn.example.org/repos/test",
            (
                '"Quotes   Too"',
                "http://svn.thirdparty.com/repos/%22Quotes%20%20%20Too%22",
                None,
                False,
            ),
        ),
        # edge cases
        (
            '-r1 http://svn.thirdparty.com/repos/test "trunk/PluginFramework"',
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("trunk/PluginFramework", "http://svn.thirdparty.com/repos/test", 1, False),
        ),
        (
            "external -r 9 http://svn.thirdparty.com/repos/test",
            "tags",
            "http://svn.example.org/repos/test",
            ("external", "http://svn.thirdparty.com/repos/test", 9, False),
        ),
        (
            "./external http://svn.thirdparty.com/repos/test",
            "tags",
            "http://svn.example.org/repos/test",
            ("external", "http://svn.thirdparty.com/repos/test", None, False),
        ),
        (
            ".external http://svn.thirdparty.com/repos/test",
            "tags",
            "http://svn.example.org/repos/test",
            (".external", "http://svn.thirdparty.com/repos/test", None, False),
        ),
        (
            "external/ http://svn.thirdparty.com/repos/test",
            "tags",
            "http://svn.example.org/repos/test",
            ("external", "http://svn.thirdparty.com/repos/test", None, False),
        ),
        (
            "external ttp://svn.thirdparty.com/repos/test",
            "tags",
            "http://svn.example.org/repos/test",
            ("external", "ttp://svn.thirdparty.com/repos/test", None, False),
        ),
        (
            "external http//svn.thirdparty.com/repos/test",
            "tags",
            "http://svn.example.org/repos/test",
            ("external", "http//svn.thirdparty.com/repos/test", None, False),
        ),
        (
            "C:\\code\\repo\\external http://svn.thirdparty.com/repos/test",
            "tags",
            "http://svn.example.org/repos/test",
            ("C:coderepoexternal", "http://svn.thirdparty.com/repos/test", None, False),
        ),
        (
            "C:\\\\code\\\\repo\\\\external http://svn.thirdparty.com/repos/test",
            "tags",
            "http://svn.example.org/repos/test",
            (
                "C:\\code\\repo\\external",
                "http://svn.thirdparty.com/repos/test",
                None,
                False,
            ),
        ),
        (
            "-r 123 http://svn.example.com/repos/sounds@100 third-party/sounds",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("third-party/sounds", "http://svn.example.com/repos/sounds", 123, False),
        ),
        (
            "-r 123 http://svn.example.com/repos/sounds@150 third-party/sounds",
            "trunk/externals",
            "http://svn.example.org/repos/test",
            ("third-party/sounds", "http://svn.example.com/repos/sounds", 123, False),
        ),
    ],
)
def test_parse_external_definition(external, dir_path, repo_url, expected_result):
    assert (
        utils.parse_external_definition(external, dir_path, repo_url) == expected_result
    )



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

import pytest
from subvertpy import SubversionException
from subvertpy.ra import Auth, RemoteAccess, get_username_provider

from swh.loader.svn.svn import SvnRepo
from swh.loader.svn.svn_retry import SVN_RETRY_MAX_ATTEMPTS, SVN_RETRY_WAIT_EXP_BASE
from swh.loader.tests import prepare_repository_from_archive


def _get_repo_url(archive_name, datadir, tmp_path):
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    return prepare_repository_from_archive(archive_path, "pkg-gourmet", tmp_path)


@pytest.fixture()
def sample_repo_url(datadir, tmp_path):
    return _get_repo_url("pkg-gourmet", datadir, tmp_path)


@pytest.fixture()
def sample_repo_with_externals_url(datadir, tmp_path):
    return _get_repo_url("pkg-gourmet-with-external-id", datadir, tmp_path)


class SVNClientWrapper:
    """Methods of subvertpy.client.Client cannot be patched by mocker fixture
    as they are read only attributes due to subvertpy.client module being
    a C extension module. So we use that wrapper class instead to simulate
    mocking behavior.
    """

    def __init__(self, client, exception, nb_failed_calls):
        self.client = client
        self.exception = exception
        self.nb_failed_calls = nb_failed_calls
        self.nb_calls = 0

    def _wrapped_svn_cmd(self, svn_cmd, *args, **kwargs):
        self.nb_calls = self.nb_calls + 1
        if self.nb_calls <= self.nb_failed_calls:
            raise self.exception
        else:
            return svn_cmd(*args, **kwargs)

    def export(self, *args, **kwargs):
        return self._wrapped_svn_cmd(self.client.export, *args, **kwargs)

    def checkout(self, *args, **kwargs):
        return self._wrapped_svn_cmd(self.client.checkout, *args, **kwargs)

    def propget(self, *args, **kwargs):
        return self._wrapped_svn_cmd(self.client.propget, *args, **kwargs)

    def info(self, *args, **kwargs):
        return self._wrapped_svn_cmd(self.client.info, *args, **kwargs)


class SVNRemoteAccessWrapper:
    """Methods of subvertpy.ra.RemoteAccess cannot be patched by mocker fixture
    as they are read only attributes due to subvertpy._ra module being
    a C extension module. So we use that wrapper class instead to simulate
    mocking behavior.
    """

    def __init__(self, svn_ra, exception, nb_failed_calls):
        self.svn_ra = svn_ra
        self.exception = exception
        self.nb_failed_calls = nb_failed_calls
        self.nb_calls = 0

    def _wrapped_svn_ra_cmd(self, svn_ra_cmd, *args, **kwargs):
        self.nb_calls = self.nb_calls + 1
        if self.nb_calls <= self.nb_failed_calls:
            raise self.exception
        else:
            return svn_ra_cmd(*args, **kwargs)

    def iter_log(self, *args, **kwargs):
        return self._wrapped_svn_ra_cmd(self.svn_ra.iter_log, *args, **kwargs)


def assert_sleep_calls(mock_sleep, mocker, nb_failures):
    mock_sleep.assert_has_calls(
        [
            mocker.call(param)
            for param in [SVN_RETRY_WAIT_EXP_BASE**i for i in range(nb_failures)]
        ]
    )


RETRYABLE_EXCEPTIONS = [
    SubversionException(
        "Error running context: The server unexpectedly closed the connection.",
        120108,
    ),
    SubversionException("Connection timed out", 175012),
    SubversionException("Unable to connect to a repository at URL", 170013),
    SubversionException(
        "ra_serf: The server sent a truncated HTTP response body.", 120106
    ),
    ConnectionResetError(),
    TimeoutError(),
]


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_export_retry_success(
    mocker, tmp_path, sample_repo_url, exception_to_retry
):
    svnrepo = SvnRepo(
        sample_repo_url, sample_repo_url, tmp_path, max_content_length=100000
    )

    mock_sleep = mocker.patch.object(svnrepo.export.retry, "sleep")

    nb_failed_calls = 2
    svnrepo.client = SVNClientWrapper(
        svnrepo.client, exception_to_retry, nb_failed_calls
    )

    export_path = os.path.join(tmp_path, "export")
    svnrepo.export(sample_repo_url, export_path)
    assert os.path.exists(export_path)

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_export_retry_failure(
    mocker, tmp_path, sample_repo_url, exception_to_retry
):
    svnrepo = SvnRepo(
        sample_repo_url, sample_repo_url, tmp_path, max_content_length=100000
    )

    mock_sleep = mocker.patch.object(svnrepo.export.retry, "sleep")

    nb_failed_calls = SVN_RETRY_MAX_ATTEMPTS
    svnrepo.client = SVNClientWrapper(
        svnrepo.client, exception_to_retry, nb_failed_calls
    )

    with pytest.raises(type(exception_to_retry)):
        export_path = os.path.join(tmp_path, "export")
        svnrepo.export(sample_repo_url, export_path)

    assert not os.path.exists(export_path)

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls - 1)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_checkout_retry_success(
    mocker, tmp_path, sample_repo_url, exception_to_retry
):
    svnrepo = SvnRepo(
        sample_repo_url, sample_repo_url, tmp_path, max_content_length=100000
    )

    mock_sleep = mocker.patch.object(svnrepo.checkout.retry, "sleep")

    nb_failed_calls = 2
    svnrepo.client = SVNClientWrapper(
        svnrepo.client, exception_to_retry, nb_failed_calls
    )

    checkout_path = os.path.join(tmp_path, "checkout")
    svnrepo.checkout(sample_repo_url, checkout_path, svnrepo.head_revision())
    assert os.path.exists(checkout_path)

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_checkout_retry_failure(
    mocker, tmp_path, sample_repo_url, exception_to_retry
):
    svnrepo = SvnRepo(
        sample_repo_url, sample_repo_url, tmp_path, max_content_length=100000
    )

    mock_sleep = mocker.patch.object(svnrepo.checkout.retry, "sleep")

    nb_failed_calls = SVN_RETRY_MAX_ATTEMPTS
    svnrepo.client = SVNClientWrapper(
        svnrepo.client, exception_to_retry, nb_failed_calls
    )

    checkout_path = os.path.join(tmp_path, "checkout")
    with pytest.raises(type(exception_to_retry)):
        svnrepo.checkout(sample_repo_url, checkout_path, svnrepo.head_revision())

    assert not os.path.exists(checkout_path)

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls - 1)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_propget_retry_success(
    mocker, tmp_path, sample_repo_with_externals_url, exception_to_retry
):
    svnrepo = SvnRepo(
        sample_repo_with_externals_url,
        sample_repo_with_externals_url,
        tmp_path,
        max_content_length=100000,
    )

    checkout_path = os.path.join(tmp_path, "checkout")
    svnrepo.checkout(
        sample_repo_with_externals_url,
        checkout_path,
        svnrepo.head_revision(),
        ignore_externals=True,
    )

    mock_sleep = mocker.patch.object(svnrepo.propget.retry, "sleep")

    nb_failed_calls = 2
    svnrepo.client = SVNClientWrapper(
        svnrepo.client, exception_to_retry, nb_failed_calls
    )

    externals = svnrepo.propget("svn:externals", checkout_path, None, None, True)

    assert externals

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_propget_retry_failure(
    mocker, tmp_path, sample_repo_with_externals_url, exception_to_retry
):
    svnrepo = SvnRepo(
        sample_repo_with_externals_url,
        sample_repo_with_externals_url,
        tmp_path,
        max_content_length=100000,
    )

    checkout_path = os.path.join(tmp_path, "checkout")
    svnrepo.checkout(
        sample_repo_with_externals_url,
        checkout_path,
        svnrepo.head_revision(),
        ignore_externals=True,
    )

    mock_sleep = mocker.patch.object(svnrepo.propget.retry, "sleep")

    nb_failed_calls = SVN_RETRY_MAX_ATTEMPTS
    svnrepo.client = SVNClientWrapper(
        svnrepo.client, exception_to_retry, nb_failed_calls
    )

    with pytest.raises(type(exception_to_retry)):
        svnrepo.propget("svn:externals", checkout_path, None, None, True)

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls - 1)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_remote_access_retry_success(
    mocker, tmp_path, sample_repo_url, exception_to_retry
):

    nb_failed_calls = 2
    mock_ra = mocker.patch("swh.loader.svn.svn.RemoteAccess")
    remote_access = RemoteAccess(sample_repo_url, auth=Auth([get_username_provider()]))
    mock_ra.side_effect = (
        [exception_to_retry] * nb_failed_calls
        + [remote_access]
        + [exception_to_retry] * nb_failed_calls
        + [remote_access]
    )

    mock_sleep = mocker.patch.object(SvnRepo.remote_access.retry, "sleep")

    SvnRepo(
        sample_repo_url,
        sample_repo_url,
        tmp_path,
        max_content_length=100000,
    )

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_remote_access_retry_failure(
    mocker, tmp_path, sample_repo_url, exception_to_retry
):

    nb_failed_calls = SVN_RETRY_MAX_ATTEMPTS
    mock_ra = mocker.patch("swh.loader.svn.svn.RemoteAccess")
    remote_access = RemoteAccess(sample_repo_url, auth=Auth([get_username_provider()]))
    mock_ra.side_effect = (
        [exception_to_retry] * nb_failed_calls
        + [remote_access]
        + [exception_to_retry] * nb_failed_calls
        + [remote_access]
    )

    mock_sleep = mocker.patch.object(SvnRepo.remote_access.retry, "sleep")

    with pytest.raises(type(exception_to_retry)):
        SvnRepo(
            sample_repo_url,
            sample_repo_url,
            tmp_path,
            max_content_length=100000,
        )

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls - 1)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_info_retry_success(mocker, tmp_path, sample_repo_url, exception_to_retry):
    svnrepo = SvnRepo(
        sample_repo_url, sample_repo_url, tmp_path, max_content_length=100000
    )

    mock_sleep = mocker.patch.object(svnrepo.info.retry, "sleep")

    nb_failed_calls = 2
    svnrepo.client = SVNClientWrapper(
        svnrepo.client, exception_to_retry, nb_failed_calls
    )

    info = svnrepo.info(sample_repo_url)
    assert info

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_info_retry_failure(mocker, tmp_path, sample_repo_url, exception_to_retry):
    svnrepo = SvnRepo(
        sample_repo_url, sample_repo_url, tmp_path, max_content_length=100000
    )

    mock_sleep = mocker.patch.object(svnrepo.info.retry, "sleep")

    nb_failed_calls = SVN_RETRY_MAX_ATTEMPTS
    svnrepo.client = SVNClientWrapper(
        svnrepo.client, exception_to_retry, nb_failed_calls
    )

    with pytest.raises(type(exception_to_retry)):
        svnrepo.info(sample_repo_url)

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls - 1)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_commit_info_retry_success(
    mocker, tmp_path, sample_repo_url, exception_to_retry
):
    svnrepo = SvnRepo(
        sample_repo_url, sample_repo_url, tmp_path, max_content_length=100000
    )

    mock_sleep = mocker.patch.object(svnrepo.commit_info.retry, "sleep")

    nb_failed_calls = 2
    svnrepo.conn_log = SVNRemoteAccessWrapper(
        svnrepo.conn_log, exception_to_retry, nb_failed_calls
    )

    commit = svnrepo.commit_info(revision=1)
    assert commit

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls)


@pytest.mark.parametrize("exception_to_retry", RETRYABLE_EXCEPTIONS)
def test_svn_commit_info_retry_failure(
    mocker, tmp_path, sample_repo_url, exception_to_retry
):
    svnrepo = SvnRepo(
        sample_repo_url, sample_repo_url, tmp_path, max_content_length=100000
    )

    mock_sleep = mocker.patch.object(svnrepo.commit_info.retry, "sleep")

    nb_failed_calls = SVN_RETRY_MAX_ATTEMPTS
    svnrepo.conn_log = SVNRemoteAccessWrapper(
        svnrepo.conn_log, exception_to_retry, nb_failed_calls
    )

    with pytest.raises(type(exception_to_retry)):
        svnrepo.commit_info(sample_repo_url)

    assert_sleep_calls(mock_sleep, mocker, nb_failed_calls - 1)



# Copyright (C) 2016-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import shutil
import subprocess
import textwrap
from typing import Any, Dict

import pytest
from subvertpy import SubversionException

from swh.loader.svn.loader import (
    SvnLoader,
    SvnLoaderFromDumpArchive,
    SvnLoaderFromRemoteDump,
)
from swh.loader.svn.svn import SvnRepo
from swh.loader.svn.utils import init_svn_repo_from_dump
from swh.loader.tests import (
    assert_last_visit_matches,
    check_snapshot,
    get_stats,
    prepare_repository_from_archive,
)
from swh.model.from_disk import DentryPerms
from swh.model.hashutil import hash_to_bytes
from swh.model.model import Snapshot, SnapshotBranch, TargetType

from .utils import CommitChange, CommitChangeType, add_commit

GOURMET_SNAPSHOT = Snapshot(
    id=hash_to_bytes("889cacc2731e3312abfb2b1a0c18ade82a949e07"),
    branches={
        b"HEAD": SnapshotBranch(
            target=hash_to_bytes("4876cb10aec6f708f7466dddf547567b65f6c39c"),
            target_type=TargetType.REVISION,
        )
    },
)

GOURMET_UPDATES_SNAPSHOT = Snapshot(
    id=hash_to_bytes("11086d15317014e43d2438b7ffc712c44f1b8afe"),
    branches={
        b"HEAD": SnapshotBranch(
            target=hash_to_bytes("171dc35522bfd17dda4e90a542a0377fb2fc707a"),
            target_type=TargetType.REVISION,
        )
    },
)


def test_loader_svn_not_found_no_mock(swh_storage, tmp_path):
    """Given an unknown repository, the loader visit ends up in status not_found"""
    repo_url = "unknown-repository"
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "uneventful"}

    assert_last_visit_matches(
        swh_storage,
        repo_url,
        status="not_found",
        type="svn",
    )


@pytest.mark.parametrize(
    "exception_msg",
    [
        "Unable to connect to a repository at URL",
        "Unknown URL type",
    ],
)
def test_loader_svn_not_found(swh_storage, tmp_path, exception_msg, mocker):
    """Given unknown repository issues, the loader visit ends up in status not_found"""
    mock = mocker.patch("swh.loader.svn.loader.SvnRepo")
    mock.side_effect = SubversionException(exception_msg, 0)

    unknown_repo_url = "unknown-repository"
    loader = SvnLoader(swh_storage, unknown_repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "uneventful"}

    assert_last_visit_matches(
        swh_storage,
        unknown_repo_url,
        status="not_found",
        type="svn",
    )


@pytest.mark.parametrize(
    "exception",
    [
        SubversionException("Irrelevant message, considered a failure", 10),
        SubversionException("Present but fails to read, considered a failure", 20),
        ValueError("considered a failure"),
    ],
)
def test_loader_svn_failures(swh_storage, tmp_path, exception, mocker):
    """Given any errors raised, the loader visit ends up in status failed"""
    mock = mocker.patch("swh.loader.svn.loader.SvnRepo")
    mock.side_effect = exception

    existing_repo_url = "existing-repo-url"
    loader = SvnLoader(swh_storage, existing_repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "failed"}

    assert_last_visit_matches(
        swh_storage,
        existing_repo_url,
        status="failed",
        type="svn",
    )


def test_loader_svnrdump_not_found(swh_storage, tmp_path, mocker):
    """Loading from remote dump which does not exist should end up as not_found visit"""
    unknown_repo_url = "file:///tmp/svn.code.sf.net/p/white-rats-studios/svn"

    loader = SvnLoaderFromRemoteDump(
        swh_storage, unknown_repo_url, temp_directory=tmp_path
    )

    assert loader.load() == {"status": "uneventful"}

    assert_last_visit_matches(
        swh_storage,
        unknown_repo_url,
        status="not_found",
        type="svn",
    )


def test_loader_svnrdump_no_such_revision(swh_storage, tmp_path, datadir):
    """Visit multiple times an origin with the remote loader should not raise.

    It used to fail the ingestion on the second visit with a "No such revision x,
    160006" message.

    """
    archive_ori_dump = os.path.join(datadir, "penguinsdbtools2018.dump.gz")
    archive_dump_dir = os.path.join(tmp_path, "dump")
    os.mkdir(archive_dump_dir)
    archive_dump = os.path.join(archive_dump_dir, "penguinsdbtools2018.dump.gz")
    # loader now drops the dump as soon as it's mounted so we need to make a copy first
    shutil.copyfile(archive_ori_dump, archive_dump)

    loading_path = str(tmp_path / "loading")
    os.mkdir(loading_path)

    # Prepare the dump as a local svn repository for test purposes
    temp_dir, repo_path = init_svn_repo_from_dump(
        archive_dump, root_dir=tmp_path, gzip=True
    )
    repo_url = f"file://{repo_path}"

    loader = SvnLoaderFromRemoteDump(swh_storage, repo_url, temp_directory=loading_path)
    assert loader.load() == {"status": "eventful"}
    actual_visit = assert_last_visit_matches(
        swh_storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    loader2 = SvnLoaderFromRemoteDump(
        swh_storage, repo_url, temp_directory=loading_path
    )
    # Visiting a second time the same repository should be uneventful...
    assert loader2.load() == {"status": "uneventful"}
    actual_visit2 = assert_last_visit_matches(
        swh_storage,
        repo_url,
        status="full",
        type="svn",
    )

    assert actual_visit.snapshot is not None
    # ... with the same snapshot as the first visit
    assert actual_visit2.snapshot == actual_visit.snapshot


def test_loader_svn_new_visit(swh_storage, datadir, tmp_path):
    """Eventful visit should yield 1 snapshot"""
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )
    check_snapshot(loader.snapshot, loader.storage)

    stats = get_stats(loader.storage)
    assert stats == {
        "content": 19,
        "directory": 17,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 6,
        "skipped_content": 0,
        "snapshot": 1,
    }

    check_snapshot(GOURMET_SNAPSHOT, loader.storage)


def test_loader_svn_2_visits_no_change(swh_storage, datadir, tmp_path):
    """Visit multiple times a repository with no change should yield the same snapshot"""
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}
    visit_status1 = assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )
    check_snapshot(loader.snapshot, loader.storage)

    assert loader.load() == {"status": "uneventful"}
    visit_status2 = assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )

    assert visit_status1.date < visit_status2.date
    assert visit_status1.snapshot == visit_status2.snapshot

    stats = get_stats(loader.storage)
    assert stats["origin_visit"] == 1 + 1  # computed twice the same snapshot
    assert stats["snapshot"] == 1

    # even starting from previous revision...
    start_revision = loader.storage.revision_get(
        [hash_to_bytes("95edacc8848369d6fb1608e887d6d2474fd5224f")]
    )[0]
    assert start_revision is not None

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)
    assert loader.load() == {"status": "uneventful"}

    stats = get_stats(loader.storage)
    assert stats["origin_visit"] == 2 + 1
    # ... with no change in repository, this yields the same snapshot
    assert stats["snapshot"] == 1

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )


def test_loader_tampered_repository(swh_storage, datadir, tmp_path):
    """In this scenario, the dump has been tampered with to modify the
    commit log [1].  This results in a hash divergence which is
    detected at startup after a new run for the same origin.

    In effect, this will perform a complete reloading of the repository.

    [1] Tampering with revision 6 log message following:

    ```
     tar xvf pkg-gourmet.tgz  # initial repository ingested
     cd pkg-gourmet/
     echo "Tampering with commit log message for fun and profit" > log.txt
     svnadmin setlog . -r 6 log.txt --bypass-hooks
     tar cvf pkg-gourmet-tampered-rev6-log.tgz pkg-gourmet/
    ```
    """
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)
    assert loader.load() == {"status": "eventful"}
    check_snapshot(GOURMET_SNAPSHOT, loader.storage)

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )
    check_snapshot(loader.snapshot, loader.storage)

    archive_path2 = os.path.join(datadir, "pkg-gourmet-tampered-rev6-log.tgz")
    repo_tampered_url = prepare_repository_from_archive(
        archive_path2, archive_name, tmp_path
    )

    loader2 = SvnLoader(
        swh_storage, repo_tampered_url, origin_url=repo_url, temp_directory=tmp_path
    )
    assert loader2.load() == {"status": "eventful"}

    assert_last_visit_matches(
        loader2.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=hash_to_bytes("5aa61959e788e281fd6e187053d0f46c68e8d8bb"),
    )
    check_snapshot(loader.snapshot, loader.storage)

    stats = get_stats(loader.storage)
    assert stats["origin"] == 1
    assert stats["origin_visit"] == 2
    assert stats["snapshot"] == 2


def test_loader_svn_visit_with_changes(swh_storage, datadir, tmp_path):
    """In this scenario, the repository has been updated with new changes.
    The loading visit should result in new objects stored and 1 new
    snapshot.

    """
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_initial_url = prepare_repository_from_archive(
        archive_path, archive_name, tmp_path
    )

    # repo_initial_url becomes the origin_url we want to visit some more below
    loader = SvnLoader(swh_storage, repo_initial_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}
    visit_status1 = assert_last_visit_matches(
        loader.storage,
        repo_initial_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )
    check_snapshot(GOURMET_SNAPSHOT, loader.storage)

    archive_path = os.path.join(datadir, "pkg-gourmet-with-updates.tgz")
    repo_updated_url = prepare_repository_from_archive(
        archive_path, "pkg-gourmet", tmp_path
    )

    loader = SvnLoader(
        swh_storage,
        repo_updated_url,
        origin_url=repo_initial_url,
        temp_directory=tmp_path,
    )

    assert loader.load() == {"status": "eventful"}
    visit_status2 = assert_last_visit_matches(
        loader.storage,
        repo_updated_url,
        status="full",
        type="svn",
        snapshot=GOURMET_UPDATES_SNAPSHOT.id,
    )

    assert visit_status1.date < visit_status2.date
    assert visit_status1.snapshot != visit_status2.snapshot

    stats = get_stats(loader.storage)
    assert stats == {
        "content": 22,
        "directory": 28,
        "origin": 1,
        "origin_visit": 2,
        "release": 0,
        "revision": 11,
        "skipped_content": 0,
        "snapshot": 2,
    }

    check_snapshot(GOURMET_UPDATES_SNAPSHOT, loader.storage)

    # Let's start the ingestion from the start, this should yield the same result
    loader = SvnLoader(
        swh_storage,
        repo_updated_url,
        origin_url=repo_initial_url,
        incremental=False,
        temp_directory=tmp_path,
    )
    assert loader.load() == {"status": "eventful"}
    visit_status3 = assert_last_visit_matches(
        loader.storage,
        repo_updated_url,
        status="full",
        type="svn",
        snapshot=GOURMET_UPDATES_SNAPSHOT.id,
    )
    assert visit_status2.date < visit_status3.date
    assert visit_status3.snapshot == visit_status2.snapshot
    check_snapshot(GOURMET_UPDATES_SNAPSHOT, loader.storage)

    stats = get_stats(loader.storage)
    assert stats["origin"] == 1  # always the same visit
    assert stats["origin_visit"] == 2 + 1  # 1 more visit
    assert stats["snapshot"] == 2  # no new snapshot


def test_loader_svn_visit_start_from_revision(swh_storage, datadir, tmp_path):
    """Starting from existing revision, next visit on changed repo should yield 1 new
    snapshot.

    """
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_initial_url = prepare_repository_from_archive(
        archive_path, archive_name, tmp_path
    )

    # repo_initial_url becomes the origin_url we want to visit some more below
    loader = SvnLoader(swh_storage, repo_initial_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}
    visit_status1 = assert_last_visit_matches(
        loader.storage,
        repo_initial_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )
    check_snapshot(GOURMET_SNAPSHOT, loader.storage)

    start_revision = loader.storage.revision_get(
        [hash_to_bytes("95edacc8848369d6fb1608e887d6d2474fd5224f")]
    )[0]
    assert start_revision is not None

    archive_path = os.path.join(datadir, "pkg-gourmet-with-updates.tgz")
    repo_updated_url = prepare_repository_from_archive(
        archive_path, "pkg-gourmet", tmp_path
    )

    # we'll start from start_revision
    loader = SvnLoader(
        swh_storage,
        repo_updated_url,
        origin_url=repo_initial_url,
        temp_directory=tmp_path,
    )

    assert loader.load() == {"status": "eventful"}

    # nonetheless, we obtain the same snapshot (as previous tests on that repository)
    visit_status2 = assert_last_visit_matches(
        loader.storage,
        repo_updated_url,
        status="full",
        type="svn",
        snapshot=GOURMET_UPDATES_SNAPSHOT.id,
    )

    assert visit_status1.date < visit_status2.date
    assert visit_status1.snapshot != visit_status2.snapshot

    stats = get_stats(loader.storage)
    assert stats == {
        "content": 22,
        "directory": 28,
        "origin": 1,
        "origin_visit": 2,
        "release": 0,
        "revision": 11,
        "skipped_content": 0,
        "snapshot": 2,
    }

    check_snapshot(GOURMET_UPDATES_SNAPSHOT, loader.storage)


def test_loader_svn_visit_with_eol_style(swh_storage, datadir, tmp_path):
    """Check that a svn repo containing a versioned file with CRLF line
    endings with svn:eol-style property set to 'native' (this is a
    violation of svn specification as the file should have been
    stored with LF line endings) can be loaded anyway.

    """
    archive_name = "mediawiki-repo-r407-eol-native-crlf"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}
    mediawiki_snapshot = Snapshot(
        id=hash_to_bytes("d6d6e9703f157c5702d9a4a5dec878926ed4ab76"),
        branches={
            b"HEAD": SnapshotBranch(
                target=hash_to_bytes("7da4975c363101b819756d33459f30a866d01b1b"),
                target_type=TargetType.REVISION,
            )
        },
    )
    check_snapshot(mediawiki_snapshot, loader.storage)

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=mediawiki_snapshot.id,
    )

    stats = get_stats(loader.storage)
    assert stats["origin"] == 1
    assert stats["origin_visit"] == 1
    assert stats["snapshot"] == 1


def test_loader_svn_visit_with_mixed_crlf_lf(swh_storage, datadir, tmp_path):
    """Check that a svn repo containing a versioned file with mixed
    CRLF/LF line endings with svn:eol-style property set to 'native'
    (this is a violation of svn specification as mixed line endings
    for textual content should not be stored when the svn:eol-style
    property is set) can be loaded anyway.

    """
    archive_name = "pyang-repo-r343-eol-native-mixed-lf-crlf"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}
    pyang_snapshot = Snapshot(
        id=hash_to_bytes("6d9590de11b00a5801de0ff3297c5b44bbbf7d24"),
        branches={
            b"HEAD": SnapshotBranch(
                target=hash_to_bytes("9c6962eeb9164a636c374be700672355e34a98a7"),
                target_type=TargetType.REVISION,
            )
        },
    )
    check_snapshot(pyang_snapshot, loader.storage)

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=pyang_snapshot.id,
    )

    stats = get_stats(loader.storage)
    assert stats["origin"] == 1
    assert stats["origin_visit"] == 1
    assert stats["snapshot"] == 1


def test_loader_svn_with_symlink(swh_storage, datadir, tmp_path):
    """Repository with symlinks should be ingested ok

    Edge case:
       - first create a file and commit it.
         Remove it, then add folder holding the same name, commit.
       - do the same scenario with symbolic link (instead of file)

    """
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(
        datadir, "pkg-gourmet-with-edge-case-links-and-files.tgz"
    )
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}
    gourmet_edge_cases_snapshot = Snapshot(
        id=hash_to_bytes("18e60982fe521a2546ab8c3c73a535d80462d9d0"),
        branches={
            b"HEAD": SnapshotBranch(
                target=hash_to_bytes("3f43af2578fccf18b0d4198e48563da7929dc608"),
                target_type=TargetType.REVISION,
            )
        },
    )
    check_snapshot(gourmet_edge_cases_snapshot, loader.storage)

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=gourmet_edge_cases_snapshot.id,
    )

    stats = get_stats(loader.storage)
    assert stats["origin"] == 1
    assert stats["origin_visit"] == 1
    assert stats["snapshot"] == 1
    assert stats["revision"] == 19


def test_loader_svn_with_wrong_symlinks(swh_storage, datadir, tmp_path):
    """Repository with wrong symlinks should be ingested ok nonetheless

    Edge case:
       - wrong symbolic link
       - wrong symbolic link with empty space names

    """
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, "pkg-gourmet-with-wrong-link-cases.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}
    gourmet_wrong_links_snapshot = Snapshot(
        id=hash_to_bytes("b17f38acabb90f066dedd30c29f01a02af88a5c4"),
        branches={
            b"HEAD": SnapshotBranch(
                target=hash_to_bytes("cf30d3bb9d5967d0a2bbeacc405f10a5dd9b138a"),
                target_type=TargetType.REVISION,
            )
        },
    )
    check_snapshot(gourmet_wrong_links_snapshot, loader.storage)

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=gourmet_wrong_links_snapshot.id,
    )

    stats = get_stats(loader.storage)
    assert stats["origin"] == 1
    assert stats["origin_visit"] == 1
    assert stats["snapshot"] == 1
    assert stats["revision"] == 21


def test_loader_svn_cleanup_loader(swh_storage, datadir, tmp_path):
    """Loader should clean up its working directory after the load"""
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loading_temp_directory = str(tmp_path / "loading")
    os.mkdir(loading_temp_directory)
    loader = SvnLoader(swh_storage, repo_url, temp_directory=loading_temp_directory)
    assert loader.load() == {"status": "eventful"}

    # the root temporary directory still exists
    assert os.path.exists(loader.temp_directory)
    # but it should be empty
    assert os.listdir(loader.temp_directory) == []


def test_loader_svn_cleanup_loader_from_remote_dump(swh_storage, datadir, tmp_path):
    """Loader should clean up its working directory after the load"""
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loading_temp_directory = str(tmp_path / "loading")
    os.mkdir(loading_temp_directory)

    loader = SvnLoaderFromRemoteDump(
        swh_storage, repo_url, temp_directory=loading_temp_directory
    )
    assert loader.load() == {"status": "eventful"}

    # the root temporary directory still exists
    assert os.path.exists(loader.temp_directory)
    # but it should be empty
    assert os.listdir(loader.temp_directory) == []
    # the internal temp_dir should be cleaned up though
    assert not os.path.exists(loader.temp_dir)


def test_loader_svn_cleanup_loader_from_dump_archive(swh_storage, datadir, tmp_path):
    """Loader should clean up its working directory after the load"""
    archive_ori_dump = os.path.join(datadir, "penguinsdbtools2018.dump.gz")
    archive_dump_dir = os.path.join(tmp_path, "dump")
    os.mkdir(archive_dump_dir)
    archive_dump = os.path.join(archive_dump_dir, "penguinsdbtools2018.dump.gz")
    # loader now drops the dump as soon as it's mounted so we need to make a copy first
    shutil.copyfile(archive_ori_dump, archive_dump)

    loading_path = str(tmp_path / "loading")
    os.mkdir(loading_path)

    # Prepare the dump as a local svn repository for test purposes
    temp_dir, repo_path = init_svn_repo_from_dump(
        archive_dump, root_dir=tmp_path, gzip=True
    )
    repo_url = f"file://{repo_path}"

    loader = SvnLoaderFromRemoteDump(swh_storage, repo_url, temp_directory=loading_path)
    assert loader.load() == {"status": "eventful"}

    # the root temporary directory still exists
    assert os.path.exists(loader.temp_directory)
    # but it should be empty
    assert os.listdir(loader.temp_directory) == []
    # the internal temp_dir should be cleaned up though
    assert not os.path.exists(loader.temp_dir)


def test_svn_loader_from_remote_dump(swh_storage, datadir, tmpdir_factory):
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    tmp_path = tmpdir_factory.mktemp("repo1")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loaderFromDump = SvnLoaderFromRemoteDump(
        swh_storage, repo_url, temp_directory=tmp_path
    )
    assert loaderFromDump.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loaderFromDump.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )

    # rename to another origin
    tmp_path = tmpdir_factory.mktemp("repo2")
    origin_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(
        swh_storage, repo_url, origin_url=origin_url, temp_directory=tmp_path
    )
    assert loader.load() == {"status": "eventful"}  # because are working on new origin
    assert_last_visit_matches(
        loader.storage,
        origin_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )

    check_snapshot(GOURMET_SNAPSHOT, loader.storage)

    stats = get_stats(loader.storage)
    assert stats["origin"] == 2  # created one more origin
    assert stats["origin_visit"] == 2
    assert stats["snapshot"] == 1

    loader = SvnLoader(
        swh_storage, repo_url, temp_directory=tmp_path
    )  # no change on the origin-url
    assert loader.load() == {"status": "uneventful"}
    assert_last_visit_matches(
        loader.storage,
        origin_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )

    stats = get_stats(loader.storage)
    assert stats["origin"] == 2
    assert stats["origin_visit"] == 3
    assert stats["snapshot"] == 1

    # second visit from the dump should be uneventful
    loaderFromDump = SvnLoaderFromRemoteDump(
        swh_storage, repo_url, temp_directory=tmp_path
    )
    assert loaderFromDump.load() == {"status": "uneventful"}


def test_svn_loader_from_remote_dump_incremental_load_on_stale_repo(
    swh_storage, datadir, tmp_path, mocker
):
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    # first load: a dump file will be created, mounted to a local repository
    # and the latter will be loaded into the archive
    loaderFromDump = SvnLoaderFromRemoteDump(
        swh_storage, repo_url, temp_directory=tmp_path
    )
    assert loaderFromDump.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loaderFromDump.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )

    # second load on same repository: the loader will detect there is no changes
    # since last load and will skip the dump, mount and load phases
    loaderFromDump = SvnLoaderFromRemoteDump(
        swh_storage, repo_url, temp_directory=tmp_path
    )

    loaderFromDump.dump_svn_revisions = mocker.MagicMock()
    init_svn_repo_from_dump = mocker.patch(
        "swh.loader.svn.loader.init_svn_repo_from_dump"
    )
    loaderFromDump.process_svn_revisions = mocker.MagicMock()
    loaderFromDump._check_revision_divergence = mocker.MagicMock()

    assert loaderFromDump.load() == {"status": "uneventful"}
    assert_last_visit_matches(
        loaderFromDump.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )

    # no dump
    loaderFromDump.dump_svn_revisions.assert_not_called()
    # no mount
    init_svn_repo_from_dump.assert_not_called()
    # no loading
    loaderFromDump.process_svn_revisions.assert_not_called()
    # no redundant post_load processing
    loaderFromDump._check_revision_divergence.assert_not_called()


def test_svn_loader_from_remote_dump_incremental_load_on_non_stale_repo(
    swh_storage, datadir, tmp_path, mocker
):
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    # first load
    loader = SvnLoaderFromRemoteDump(swh_storage, repo_url, temp_directory=tmp_path)
    loader.load()

    archive_path = os.path.join(datadir, "pkg-gourmet-with-updates.tgz")
    repo_updated_url = prepare_repository_from_archive(
        archive_path, archive_name, tmp_path
    )

    # second load
    loader = SvnLoaderFromRemoteDump(
        swh_storage, repo_updated_url, temp_directory=tmp_path
    )

    dump_svn_revisions = mocker.spy(loader, "dump_svn_revisions")
    process_svn_revisions = mocker.spy(loader, "process_svn_revisions")

    loader.load()

    dump_svn_revisions.assert_called()
    process_svn_revisions.assert_called()


def test_loader_user_defined_svn_properties(swh_storage, datadir, tmp_path):
    """Edge cases: The repository held some user defined svn-properties with special
    encodings, this prevented the repository from being loaded even though we do not
    ingest those information.

    """
    archive_name = "httthttt"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(swh_storage, repo_url)

    assert loader.load() == {"status": "eventful"}
    expected_snapshot = Snapshot(
        id=hash_to_bytes("70487267f682c07e52a2371061369b6cf5bffa47"),
        branches={
            b"HEAD": SnapshotBranch(
                target=hash_to_bytes("604a17dbb15e8d7ecb3e9f3768d09bf493667a93"),
                target_type=TargetType.REVISION,
            )
        },
    )
    check_snapshot(expected_snapshot, loader.storage)

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=expected_snapshot.id,
    )

    stats = get_stats(loader.storage)
    assert stats["origin"] == 1
    assert stats["origin_visit"] == 1
    assert stats["snapshot"] == 1
    assert stats["revision"] == 7


def test_loader_svn_dir_added_then_removed(swh_storage, datadir, tmp_path):
    """Loader should handle directory removal when processing a commit"""
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}-add-remove-dir.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)


def test_loader_svn_loader_from_dump_archive(swh_storage, datadir, tmp_path):
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    dump_filename = f"{archive_name}.dump"

    with open(os.path.join(tmp_path, dump_filename), "wb") as dump_file:
        # create compressed dump file of pkg-gourmet repo
        subprocess.run(["svnrdump", "dump", repo_url], stdout=dump_file)
        subprocess.run(["gzip", dump_filename], cwd=tmp_path)

        # load svn repo from that compressed dump file
        loader = SvnLoaderFromDumpArchive(
            swh_storage,
            url=repo_url,
            archive_path=os.path.join(tmp_path, f"{dump_filename}.gz"),
            temp_directory=tmp_path,
        )

        assert loader.load() == {"status": "eventful"}

        assert_last_visit_matches(
            loader.storage,
            repo_url,
            status="full",
            type="svn",
            snapshot=GOURMET_SNAPSHOT.id,
        )

        check_snapshot(GOURMET_SNAPSHOT, loader.storage)

        assert get_stats(loader.storage) == {
            "content": 19,
            "directory": 17,
            "origin": 1,
            "origin_visit": 1,
            "release": 0,
            "revision": 6,
            "skipped_content": 0,
            "snapshot": 1,
        }


def test_loader_eol_style_file_property_handling_edge_case(
    swh_storage, repo_url, tmp_path
):

    # # first commit
    add_commit(
        repo_url,
        (
            "Add a directory containing a file with CRLF end of line "
            "and set svn:eol-style property to native so CRLF will be "
            "replaced by LF in the file when exporting the revision"
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="directory/file_with_crlf_eol.txt",
                properties={"svn:eol-style": "native"},
                data=b"Hello world!\r\n",
            )
        ],
    )

    # second commit
    add_commit(
        repo_url,
        "Remove previously added directory and file",
        [
            CommitChange(
                change_type=CommitChangeType.Delete,
                path="directory/",
            )
        ],
    )

    # third commit
    add_commit(
        repo_url,
        (
            "Add again same directory containing same file with CRLF end of line "
            "but do not set svn:eol-style property value so CRLF will not be "
            "replaced by LF when exporting the revision"
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="directory/file_with_crlf_eol.txt",
                data=b"Hello world!\r\n",
            )
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    assert get_stats(loader.storage) == {
        "content": 2,
        "directory": 5,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 3,
        "skipped_content": 0,
        "snapshot": 1,
    }


def get_head_revision_paths_info(loader: SvnLoader) -> Dict[bytes, Dict[str, Any]]:
    assert loader.snapshot is not None
    root_dir = loader.snapshot.branches[b"HEAD"].target
    revision = loader.storage.revision_get([root_dir])[0]
    assert revision is not None

    paths = {}
    for entry in loader.storage.directory_ls(revision.directory, recursive=True):
        paths[entry["name"]] = entry
    return paths


def test_loader_eol_style_on_svn_link_handling(swh_storage, repo_url, tmp_path):

    # first commit
    add_commit(
        repo_url,
        (
            "Add a regular file, a directory and a link to the regular file "
            "in the directory. Set svn:eol-style property for the regular "
            "file and the link. Set svn:special property for the link."
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="file_with_crlf_eol.txt",
                properties={"svn:eol-style": "native"},
                data=b"Hello world!\r\n",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="directory/file_with_crlf_eol.txt",
                properties={"svn:eol-style": "native", "svn:special": "*"},
                data=b"link ../file_with_crlf_eol.txt",
            ),
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    # check loaded objects are those expected
    assert get_stats(loader.storage) == {
        "content": 2,
        "directory": 2,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 1,
        "skipped_content": 0,
        "snapshot": 1,
    }

    paths = get_head_revision_paths_info(loader)

    assert (
        loader.storage.content_get_data(paths[b"file_with_crlf_eol.txt"]["sha1"])
        == b"Hello world!\n"
    )

    assert paths[b"directory/file_with_crlf_eol.txt"]["perms"] == DentryPerms.symlink
    assert (
        loader.storage.content_get_data(
            paths[b"directory/file_with_crlf_eol.txt"]["sha1"]
        )
        == b"../file_with_crlf_eol.txt"
    )


def test_loader_svn_special_property_unset(swh_storage, repo_url, tmp_path):

    # first commit
    add_commit(
        repo_url,
        (
            "Create a regular file, a link to a file and a link to an "
            "external file. Set the svn:special property on the links."
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="file.txt",
                data=b"Hello world!\n",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="link.txt",
                properties={"svn:special": "*"},
                data=b"link ./file.txt",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="external_link.txt",
                properties={"svn:special": "*"},
                data=b"link /home/user/data.txt",
            ),
        ],
    )

    # second commit
    add_commit(
        repo_url,
        "Unset the svn:special property on the links.",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="link.txt",
                properties={"svn:special": None},
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="external_link.txt",
                properties={"svn:special": None},
            ),
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    # check loaded objects are those expected
    assert get_stats(loader.storage) == {
        "content": 5,
        "directory": 2,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 2,
        "skipped_content": 0,
        "snapshot": 1,
    }

    paths = get_head_revision_paths_info(loader)

    assert paths[b"link.txt"]["perms"] == DentryPerms.content
    assert (
        loader.storage.content_get_data(paths[b"link.txt"]["sha1"])
        == b"link ./file.txt"
    )

    assert paths[b"external_link.txt"]["perms"] == DentryPerms.content
    assert (
        loader.storage.content_get_data(paths[b"external_link.txt"]["sha1"])
        == b"link /home/user/data.txt"
    )


def test_loader_invalid_svn_eol_style_property_value(swh_storage, repo_url, tmp_path):

    filename = "file_with_crlf_eol.txt"
    file_content = b"Hello world!\r\n"

    # # first commit
    add_commit(
        repo_url,
        (
            "Add a file with CRLF end of line and set svn:eol-style property "
            "to an invalid value."
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path=filename,
                properties={"svn:eol-style": "foo"},
                data=file_content,
            )
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    paths = get_head_revision_paths_info(loader)
    # end of lines should not have been processed
    assert (
        loader.storage.content_get_data(paths[filename.encode()]["sha1"])
        == file_content
    )


def test_loader_first_revision_is_not_number_one(
    swh_storage, mocker, repo_url, tmp_path
):
    class SvnRepoSkipFirstRevision(SvnRepo):
        def logs(self, revision_start, revision_end):
            """Overrides logs method to skip revision number one in yielded revisions"""
            yield from super().logs(revision_start + 1, revision_end)

    from swh.loader.svn import loader

    mocker.patch.object(loader, "SvnRepo", SvnRepoSkipFirstRevision)

    for filename in ("foo", "bar", "baz"):
        add_commit(
            repo_url,
            f"Add {filename} file",
            [
                CommitChange(
                    change_type=CommitChangeType.AddOrUpdate,
                    path=filename,
                    data=f"{filename}\n".encode(),
                )
            ],
        )

    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)

    # post loading will detect an issue and make a partial visit with a snapshot
    assert loader.load() == {"status": "failed"}

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="partial",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    assert get_stats(loader.storage) == {
        "content": 2,
        "directory": 2,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 2,
        "skipped_content": 0,
        "snapshot": 1,
    }


def test_loader_svn_special_property_on_binary_file(swh_storage, repo_url, tmp_path):
    """When a file has the svn:special property set but is not a svn link,
    it might be truncated under certain conditions when performing an export
    operation."""

    data = (
        b"!<symlink>\xff\xfea\x00p\x00t\x00-\x00c\x00y\x00g\x00.\x00s\x00h\x00\x00\x00"
    )

    # first commit
    add_commit(
        repo_url,
        (
            "Add a non svn link binary file and set the svn:special property on it."
            "That file will be truncated when exporting it."
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="binary_file",
                properties={"svn:special": "*"},
                data=data,
            ),
        ],
    )

    # second commit
    add_commit(
        repo_url,
        (
            "Add a non svn link binary file and set the svn:special and "
            "svn:mime-type properties on it."
            "That file will not be truncated when exporting it."
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="another_binary_file",
                properties={
                    "svn:special": "*",
                    "svn:mime-type": "application/octet-stream",
                },
                data=data,
            ),
        ],
    )

    # third commit
    add_commit(
        repo_url,
        "Remove the svn:special property on the previously added files",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="binary_file",
                properties={"svn:special": None},
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="another_binary_file",
                properties={"svn:special": None},
            ),
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)


def test_loader_last_revision_divergence(swh_storage, datadir, tmp_path):
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    class SvnLoaderRevisionDivergence(SvnLoader):
        def _check_revision_divergence(self, count, rev, dir_id):
            raise ValueError("revision divergence detected")

    loader = SvnLoaderRevisionDivergence(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "failed"}

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="partial",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )
    check_snapshot(GOURMET_SNAPSHOT, loader.storage)


def test_loader_delete_directory_while_file_has_same_prefix(
    swh_storage, repo_url, tmp_path
):

    # first commit
    add_commit(
        repo_url,
        "Add a file and a directory with same prefix",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="foo/bar.c",
                data=b'#include "../foo.c"',
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="foo.c",
                data=b"int foo() {return 0;}",
            ),
        ],
    )

    # second commit
    add_commit(
        repo_url,
        "Delete previously added directory and update file content",
        [
            CommitChange(change_type=CommitChangeType.Delete, path="foo"),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="foo.c",
                data=b"int foo() {return 1;}",
            ),
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)


def test_svn_loader_incremental(swh_storage, repo_url, tmp_path):

    # first commit
    add_commit(
        repo_url,
        (
            "Add a directory containing a file with CRLF end of line "
            "and set svn:eol-style property to native so CRLF will be "
            "replaced by LF in the file when exporting the revision"
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="file_with_crlf_eol.txt",
                properties={"svn:eol-style": "native"},
                data=b"Hello world!\r\n",
            )
        ],
    )

    # first load
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)
    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    # second commit
    add_commit(
        repo_url,
        "Modify previously added file",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="file_with_crlf_eol.txt",
                data=b"Hello World!\r\n",
            )
        ],
    )

    # second load, incremental
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)
    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    # third commit
    add_commit(
        repo_url,
        "Unset svn:eol-style property on file",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="file_with_crlf_eol.txt",
                properties={"svn:eol-style": None},
            )
        ],
    )

    # third load, incremental
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)
    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)


def test_svn_loader_incremental_replay_start_with_empty_directory(
    swh_storage, mocker, repo_url, tmp_path
):

    # first commit
    add_commit(
        repo_url,
        ("Add a file"),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="foo.txt",
                data=b"foo\n",
            )
        ],
    )

    # first load
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)
    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    # second commit
    add_commit(
        repo_url,
        "Modify previously added file",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="foo.txt",
                data=b"bar\n",
            )
        ],
    )

    class SvnRepoCheckReplayStartWithEmptyDirectory(SvnRepo):
        def swh_hash_data_per_revision(self, start_revision: int, end_revision: int):
            """Overrides swh_hash_data_per_revision method to grab the content
            of the directory where the svn revisions will be replayed before that
            process starts."""
            self.replay_dir_content_before_start = [
                os.path.join(root, name)
                for root, _, files in os.walk(self.local_url)
                for name in files
            ]
            yield from super().swh_hash_data_per_revision(start_revision, end_revision)

    from swh.loader.svn import loader

    mocker.patch.object(loader, "SvnRepo", SvnRepoCheckReplayStartWithEmptyDirectory)

    # second load, incremental
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path)
    loader.load()

    # check work directory was empty before replaying revisions
    assert loader.svnrepo.replay_dir_content_before_start == []


def test_loader_svn_executable_property_on_svn_link_handling(
    swh_storage, repo_url, tmp_path
):

    # first commit
    add_commit(
        repo_url,
        (
            "Add an executable file and a svn link to it."
            "Set svn:executable property for both paths."
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello-world",
                properties={"svn:executable": "*"},
                data=b"#!/bin/bash\necho Hello World !",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello",
                properties={"svn:executable": "*", "svn:special": "*"},
                data=b"link hello-world",
            ),
        ],
    )

    # second commit
    add_commit(
        repo_url,
        (
            "Remove executable file, unset link and replace it with executable content."
            "As the link was previously marked as executable, execution rights should"
            "be set after turning it to a regular file."
        ),
        [
            CommitChange(change_type=CommitChangeType.Delete, path="hello-world"),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello",
                properties={"svn:special": None},
                data=b"#!/bin/bash\necho Hello World !",
            ),
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)


def test_loader_svn_add_property_on_link(swh_storage, repo_url, tmp_path):

    # first commit
    add_commit(
        repo_url,
        "Add an executable file and a svn link to it.",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello-world",
                properties={"svn:executable": "*"},
                data=b"#!/bin/bash\necho Hello World !",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello",
                properties={"svn:special": "*"},
                data=b"link hello-world",
            ),
        ],
    )

    # second commit
    add_commit(
        repo_url,
        "Set svn:eol-style property on link",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello",
                properties={"svn:eol-style": "native"},
            ),
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)


def test_loader_svn_link_parsing(swh_storage, repo_url, tmp_path):

    # first commit
    add_commit(
        repo_url,
        "Add an executable file and a svn link to it.",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello-world",
                properties={"svn:executable": "*"},
                data=b"#!/bin/bash\necho Hello World !",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello",
                properties={"svn:special": "*"},
                data=b"link hello-world",
            ),
        ],
    )

    # second commit
    add_commit(
        repo_url,
        "Update svn link content",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello",
                data=b"link hello-world\r\n",
            ),
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)


def test_loader_svn_empty_local_dir_before_post_load(swh_storage, datadir, tmp_path):
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    class SvnLoaderPostLoadLocalDirIsEmpty(SvnLoader):
        def post_load(self, success=True):
            if success:
                self.local_dirname_content = [
                    os.path.join(root, name)
                    for root, _, files in os.walk(self.svnrepo.local_dirname)
                    for name in files
                ]
            return super().post_load(success)

    loader = SvnLoaderPostLoadLocalDirIsEmpty(
        swh_storage, repo_url, temp_directory=tmp_path
    )

    assert loader.load() == {"status": "eventful"}

    assert loader.local_dirname_content == []

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )
    check_snapshot(GOURMET_SNAPSHOT, loader.storage)


def _dump_project(tmp_path, origin_url):
    svnrdump_cmd = ["svnrdump", "dump", origin_url]
    dump_path = f"{tmp_path}/repo.dump"
    with open(dump_path, "wb") as dump_file:
        subprocess.run(svnrdump_cmd, stdout=dump_file)
    subprocess.run(["gzip", dump_path])
    return dump_path + ".gz"


def test_loader_svn_add_property_on_directory_link(swh_storage, repo_url, tmp_path):

    # first commit
    add_commit(
        repo_url,
        "Add an executable file in a directory and a svn link to the directory.",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="code/hello-world",
                properties={"svn:executable": "*"},
                data=b"#!/bin/bash\necho Hello World !",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello",
                properties={"svn:special": "*"},
                data=b"link code",
            ),
        ],
    )

    # second commit
    add_commit(
        repo_url,
        "Set svn:eol-style property on link",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="hello",
                properties={"svn:eol-style": "native"},
            ),
        ],
    )

    # instantiate a svn loader checking after each processed revision that
    # the repository filesystem it reconstructed does not differ from a subversion
    # export of that revision
    loader = SvnLoader(swh_storage, repo_url, temp_directory=tmp_path, check_revision=1)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)


@pytest.mark.parametrize(
    "svn_loader_cls", [SvnLoader, SvnLoaderFromDumpArchive, SvnLoaderFromRemoteDump]
)
def test_loader_with_subprojects(swh_storage, repo_url, tmp_path, svn_loader_cls):

    # first commit
    add_commit(
        repo_url,
        "Add first project in repository",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="project1/foo.sh",
                data=b"#!/bin/bash\necho foo",
            ),
        ],
    )

    # second commit
    add_commit(
        repo_url,
        "Add second project in repository",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="project2/bar.sh",
                data=b"#!/bin/bash\necho bar",
            ),
        ],
    )

    # third commit
    add_commit(
        repo_url,
        "Add third project in repository",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="project3/baz.sh",
                data=b"#!/bin/bash\necho baz",
            ),
        ],
    )

    for i in range(1, 4):
        # load each project in the repository separately and check behavior
        # is the same if origin URL has a trailing slash or not
        origin_url = f"{repo_url}/project{i}{'/' if i%2 else ''}"

        loader_params = {
            "storage": swh_storage,
            "url": origin_url,
            "origin_url": origin_url,
            "temp_directory": tmp_path,
            "incremental": True,
            "check_revision": 1,
        }

        if svn_loader_cls == SvnLoaderFromDumpArchive:
            loader_params["archive_path"] = _dump_project(tmp_path, origin_url)

        loader = svn_loader_cls(**loader_params)

        assert loader.load() == {"status": "eventful"}
        assert_last_visit_matches(
            loader.storage,
            origin_url,
            status="full",
            type="svn",
        )
        check_snapshot(loader.snapshot, loader.storage)

        if svn_loader_cls == SvnLoaderFromDumpArchive:
            loader_params["archive_path"] = _dump_project(tmp_path, origin_url)

        loader = svn_loader_cls(**loader_params)

        assert loader.load() == {"status": "uneventful"}

        # each project origin must have
        assert get_stats(loader.storage) == {
            "content": i,  # one content
            "directory": 2 * i,  # two directories
            "origin": i,
            "origin_visit": 2 * i,  # two visits
            "release": 0,
            "revision": i,  # one revision
            "skipped_content": 0,
            "snapshot": i,  # one snapshot
        }


@pytest.mark.parametrize(
    "svn_loader_cls", [SvnLoader, SvnLoaderFromDumpArchive, SvnLoaderFromRemoteDump]
)
def test_loader_subproject_root_dir_removal(
    swh_storage, repo_url, tmp_path, svn_loader_cls
):

    # first commit
    add_commit(
        repo_url,
        "Add project in repository",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="project/foo.sh",
                data=b"#!/bin/bash\necho foo",
            ),
        ],
    )

    # second commit
    add_commit(
        repo_url,
        "Remove project root directory",
        [CommitChange(change_type=CommitChangeType.Delete, path="project/")],
    )

    # third commit
    add_commit(
        repo_url,
        "Re-add project in repository",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="project/foo.sh",
                data=b"#!/bin/bash\necho foo",
            ),
        ],
    )

    origin_url = f"{repo_url}/project"

    loader_params = {
        "storage": swh_storage,
        "url": origin_url,
        "origin_url": origin_url,
        "temp_directory": tmp_path,
        "incremental": True,
        "check_revision": 1,
    }

    if svn_loader_cls == SvnLoaderFromDumpArchive:
        loader_params["archive_path"] = _dump_project(tmp_path, origin_url)

    loader = svn_loader_cls(**loader_params)

    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        origin_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)

    if svn_loader_cls == SvnLoaderFromDumpArchive:
        loader_params["archive_path"] = _dump_project(tmp_path, origin_url)

    loader = svn_loader_cls(**loader_params)

    assert loader.load() == {"status": "uneventful"}


@pytest.mark.parametrize("svn_loader_cls", [SvnLoader, SvnLoaderFromRemoteDump])
def test_loader_svn_not_found_after_successful_visit(
    swh_storage, datadir, tmp_path, svn_loader_cls
):
    archive_name = "pkg-gourmet"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = svn_loader_cls(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "eventful"}

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
        snapshot=GOURMET_SNAPSHOT.id,
    )
    check_snapshot(loader.snapshot, loader.storage)

    # simulate removal of remote repository
    shutil.rmtree(repo_url.replace("file://", ""))

    loader = svn_loader_cls(swh_storage, repo_url, temp_directory=tmp_path)

    assert loader.load() == {"status": "uneventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="not_found",
        type="svn",
        snapshot=None,
    )


def test_loader_svn_from_remote_dump_url_redirect(swh_storage, tmp_path, mocker):
    repo_url = "http://svn.example.org/repo"
    repo_redirect_url = "https://svn.example.org/repo"

    # mock remote subversion operations
    from swh.loader.svn.svn import client

    mocker.patch("swh.loader.svn.svn.RemoteAccess")
    init_svn_repo_from_dump = mocker.patch(
        "swh.loader.svn.loader.init_svn_repo_from_dump"
    )
    init_svn_repo_from_dump.return_value = ("", "")
    mock_client = mocker.MagicMock()
    mocker.patch.object(client, "Client", mock_client)

    class Info:
        repos_root_url = repo_redirect_url
        url = repo_redirect_url

    mock_client().info.return_value = {"repo": Info()}

    # init remote dump loader and mock some methods
    loader = SvnLoaderFromRemoteDump(swh_storage, repo_url, temp_directory=tmp_path)
    loader.dump_svn_revisions = mocker.MagicMock(return_value=("", -1))
    loader.start_from = mocker.MagicMock(return_value=(0, 0))

    # prepare loading
    loader.prepare()

    # check redirection URL has been used to dump repository
    assert loader.dump_svn_revisions.call_args_list[0][0][0] == repo_redirect_url


@pytest.mark.parametrize("svn_loader_cls", [SvnLoader, SvnLoaderFromRemoteDump])
def test_loader_basic_authentication_required(
    swh_storage, repo_url, tmp_path, svn_loader_cls, svnserve
):

    # add file to empty test repo
    add_commit(
        repo_url,
        "Add project in repository",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="project/foo.sh",
                data=b"#!/bin/bash\necho foo",
            ),
        ],
    )

    # compute repo URLs that will be made available by svnserve
    repo_path = repo_url.replace("file://", "")
    repo_root = os.path.dirname(repo_path)
    repo_name = os.path.basename(repo_path)
    username = "anonymous"
    password = "anonymous"
    port = 12000
    repo_url_no_auth = f"svn://localhost:{port}/{repo_name}"
    repo_url = f"svn://{username}:{password}@localhost:{port}/{repo_name}"

    # disable anonymous access and require authentication on test repo
    with open(os.path.join(repo_path, "conf", "svnserve.conf"), "w") as svnserve_conf:
        svnserve_conf.write(
            textwrap.dedent(
                """
                [general]

                # Authentication realm of the repository.
                realm = test-repository
                password-db = passwd

                # Deny all anonymous access
                anon-access = none

                # Grant authenticated users read and write privileges
                auth-access = write
                """
            )
        )

    # add a user with read/write access on test repo
    with open(os.path.join(repo_path, "conf", "passwd"), "w") as passwd:
        passwd.write(f"[users]\n{username} = {password}")

    # execute svnserve
    svnserve(repo_root, port)

    # check loading failed with no authentication
    loader = svn_loader_cls(swh_storage, repo_url_no_auth, temp_directory=tmp_path)
    assert loader.load() == {"status": "uneventful"}

    # check loading succeeded with authentication
    loader = svn_loader_cls(swh_storage, repo_url, temp_directory=tmp_path)
    assert loader.load() == {"status": "eventful"}
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )

    check_snapshot(loader.snapshot, loader.storage)


def test_loader_with_spaces_in_svn_url(swh_storage, repo_url, tmp_path):
    filename = "file with spaces.txt"
    content = b"foo"

    add_commit(
        repo_url,
        "Add file with spaces in its name",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path=filename,
                data=content,
            ),
        ],
    )

    svnrepo = SvnRepo(repo_url, repo_url, tmp_path, max_content_length=10000)

    dest_path = f"{tmp_path}/file"
    svnrepo.export(f"{repo_url}/{filename}", to=dest_path)

    with open(dest_path, "rb") as f:
        assert f.read() == content


@pytest.mark.parametrize("svn_loader_cls", [SvnLoader, SvnLoaderFromRemoteDump])
def test_loader_repo_with_copyfrom_and_replace_operations(
    swh_storage, repo_url, tmp_path, svn_loader_cls
):
    add_commit(
        repo_url,
        "Create trunk/data folder",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="trunk/data/foo",
                data=b"foo",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="trunk/data/bar",
                data=b"bar",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="trunk/data/baz/",
            ),
        ],
    )

    add_commit(
        repo_url,
        "Create trunk/project folder",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="trunk/project/",
            ),
        ],
    )

    add_commit(
        repo_url,
        "Create trunk/project/bar as copy of trunk/data/bar from revision 1",
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="trunk/project/bar",
                copyfrom_path=repo_url + "/trunk/data/bar",
                copyfrom_rev=1,
            ),
        ],
    )

    add_commit(
        repo_url,
        (
            "Create trunk/project/data/ folder as a copy of /trunk/data from revision 1"
            " and replace the trunk/project/data/baz/ folder by a trunk/project/data/baz file"
        ),
        [
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="trunk/project/data/",
                copyfrom_path=repo_url + "/trunk/data/",
                copyfrom_rev=1,
            ),
            CommitChange(
                change_type=CommitChangeType.Delete,
                path="trunk/project/data/baz/",
            ),
            CommitChange(
                change_type=CommitChangeType.AddOrUpdate,
                path="trunk/project/data/baz",
                data=b"baz",
            ),
        ],
    )

    loader = svn_loader_cls(
        swh_storage, repo_url, temp_directory=tmp_path, check_revision=1
    )

    assert loader.load() == {"status": "eventful"}

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="svn",
    )
    check_snapshot(loader.snapshot, loader.storage)
